# ruff: noqa: F401, F403
from win32more.win32.prelude import *
from win32more.winrt import (
    AwaitableProtocol,
    ContextManagerProtocol,
    FillArray,
    Generic,
    IInspectable,
    IterableProtocol,
    IUnknown,
    K,
    MappingProtocol,
    MulticastDelegate,
    PassArray,
    ReceiveArray,
    SequenceProtocol,
    T,
    TProgress,
    TResult,
    TSender,
    Tuple,
    V,
    WinRT_String,
    event,
    winrt_activatemethod,
    winrt_classmethod,
    winrt_commethod,
    winrt_factorymethod,
    winrt_mixinmethod,
    winrt_overload,
)



import sys
import types
import uuid
from ctypes import (
    CFUNCTYPE,
    POINTER,
    Array,
    _CFuncPtr,
    addressof,
    c_bool,
    c_byte,
    c_char_p,
    c_double,
    c_float,
    c_int16,
    c_int32,
    c_int64,
    c_size_t,
    c_ssize_t,
    c_ubyte,
    c_uint16,
    c_uint32,
    c_uint64,
    c_void_p,
    c_wchar,
    c_wchar_p,
    cast,
    cdll,
    pointer,
)
from ctypes import Structure as _Structure
from ctypes import Union as _Union
from itertools import zip_longest

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, get_origin
    from typing_extensions import get_type_hints as _get_type_hints
else:
    from typing import Annotated, get_origin
    from typing import get_type_hints as _get_type_hints

if "(arm64)" in sys.version.lower():
    ARCH = "ARM64"
elif "(amd64)" in sys.version.lower():
    ARCH = "X64"
else:
    ARCH = "X86"

if sys.platform == "cygwin":
    from .cygwin import ARCH, WINFUNCTYPE, WinError, windll  # noqa: F401
else:
    from ctypes import WINFUNCTYPE, WinError, windll

Byte = c_ubyte
SByte = c_byte
Char = c_wchar
Int16 = c_int16
UInt16 = c_uint16
Int32 = c_int32
UInt32 = c_uint32
Int64 = c_int64
UInt64 = c_uint64
IntPtr = c_ssize_t
UIntPtr = c_size_t
Single = c_float
Double = c_double
Bytes = c_char_p
String = c_wchar_p
Boolean = c_bool
VoidPtr = c_void_p
Void = None


class Enum:
    pass


class NativeBitfieldAttribute:
    def __init__(self, width):
        self.width = width


class FlexibleArray:
    def __class_getitem__(cls, type_):
        return type(f"{cls.__name__}_{type_.__name__}", (cls, type_ * 0), {})

    def __iter__(self):
        raise IndexError("FlexibleArray cannot be iterated.  Use [:stop].")

    def __getitem__(self, index):
        p = pointer(self._type_.from_address(addressof(self)))
        return p[index]

    def __setitem__(self, index, value):
        p = pointer(self._type_.from_address(addressof(self)))
        p[index] = value


# FIXME: How to manage com reference count?  ContextManager style?
class ComPtr(c_void_p):
    def __init__(self, value=None, own=False, move=None):
        super().__init__()
        if move is not None:
            self.value = move.value
            self._own = getattr(move, "_own", False)
            move.value = None
            move._own = False
        else:
            self.value = value
            self._own = own

    def __del__(self):
        if self and getattr(self, "_own", False):
            self.Release()

    # overwritten in winrt module
    @classmethod
    def __commit__(cls):
        cls._hints_ = get_type_hints(cls)
        if cls._hints_["extends"] is None:
            return cls
        cls.__bases__ = (cls._hints_["extends"],)
        return cls

    # overwritten in winrt module
    def as_(self, cls):
        iid = cls._iid_
        instance = cls(own=True)
        hr = self.QueryInterface(pointer(iid), pointer(instance))
        if FAILED(hr):
            raise WinError(hr)
        return instance

    def __eq__(self, other):
        return isinstance(other, ComPtr) and self.value == other.value


def _struct_union_commit(cls, start=True):
    if start:
        # First, register cls for circular reference.
        module = sys.modules[cls.__module__]
        setattr(module, cls.__name__, cls)

    hints = get_type_hints(cls)

    cls._generic_types = {}

    for name, type_ in hints.items():
        if getattr(type_, "_classid_", None) == "Windows.Foundation.IReference":
            cls._generic_types[name] = type_
            hints[name] = get_origin(type_)

    for type_ in hints.values():
        if issubclass(type_, (Structure, Union)) and "_fields_" not in type_.__dict__:
            # nested struct or circular reference
            _struct_union_commit(type_, False)

    # <3.13: hasattr(cls, "_fields_") not work for Union.
    if "_fields_" in cls.__dict__:
        raise CircularReferenceResolved()

    bitfields = {}
    for name, type_ in get_type_hints(cls, include_extras=True).items():
        if get_origin(type_) is Annotated:
            for metadata in type_.__metadata__:
                if isinstance(metadata, NativeBitfieldAttribute):
                    bitfields[name] = metadata.width

    fields = []
    for name, type_ in hints.items():
        if name in bitfields:
            fields.append((name, type_, bitfields[name]))
        else:
            fields.append((name, type_))

    cls._fields_ = fields

    _hook_descriptor(cls, cls)

    return cls


def _hook_descriptor(cls, struct):
    anonymous = set(getattr(struct, "_anonymous_", []))
    for name, type_, *_ in struct._fields_:
        if name in anonymous:
            _hook_descriptor(cls, type_)
            continue
        if issubclass(type_, (c_char_p, c_wchar_p)):
            setattr(cls, f"{name}_as_intptr", AsIntPtrDescriptor(cls.__dict__[name]))
        if getattr(type_, "_classid_", None) == "Windows.Foundation.IReference":
            setattr(cls, name, IReferenceDescriptor(cls.__dict__[name], struct._generic_types[name]))
        else:
            # use __dict__[name] to avoid calling descriptor.__get__().
            setattr(cls, name, EasyCastDescriptor(cls.__dict__[name], type_))


class Structure(_Structure):
    __commit__ = classmethod(_struct_union_commit)


class Union(_Union):
    __commit__ = classmethod(_struct_union_commit)


class EasyCastDescriptor:
    def __init__(self, original_descriptor, type_):
        self._original_descriptor = original_descriptor
        self._type = type_

    def __get__(self, instance, owner=None):
        if instance is None:
            # _ctypes/stgdict.c:MakeFields() raises TypeError for non _ctypes.CField type.
            return self._original_descriptor.__get__(instance, owner)
        return self._original_descriptor.__get__(instance, owner)

    def __set__(self, instance, value):
        self._original_descriptor.__set__(instance, easycast(value, self._type))

    @property
    def offset(self):
        return self._original_descriptor.offset


class AsIntPtrDescriptor:
    def __init__(self, original_descriptor):
        self._original_descriptor = original_descriptor

    def __get__(self, instance, owner=None):
        if instance is None:
            return self._original_descriptor.__get__(instance, owner)
        return UIntPtr.from_buffer(instance, self._original_descriptor.offset).value

    def __set__(self, instance, value):
        self._original_descriptor.__set__(instance, value)


class IReferenceDescriptor:
    def __init__(self, original_descriptor, type_):
        self._original_descriptor = original_descriptor
        self._type = type_

    def __get__(self, instance, owner=None):
        if instance is None:
            return self._original_descriptor.__get__(instance, owner)
        return self._type(move=self._original_descriptor.__get__(instance, owner))

    def __set__(self, instance, value):
        self._original_descriptor.__set__(instance, value)

    @property
    def offset(self):
        return self._original_descriptor.offset


easycast_keep_reference = []


def easycast(obj, type_):
    if issubclass(type_, (POINTER(Int16), POINTER(UInt16))):
        if isinstance(obj, str):
            return cast(c_wchar_p(obj), type_)
        elif isinstance(obj, c_wchar_p):
            return cast(obj, type_)
    elif issubclass(type_, (POINTER(POINTER(Int16)), POINTER(POINTER(UInt16)))):
        if isinstance(obj, c_wchar_p):
            return cast(pointer(obj), type_)
    elif issubclass(type_, (c_char_p, c_wchar_p)):
        if isinstance(obj, int):
            # function doesn't support this conversion, though struct does it.
            return type_(obj)
    elif issubclass(type_, ComPtr):
        if isinstance(obj, (str, bytes)):
            # disable default conversion
            raise TypeError(f"cannot convert from {type(obj)} to ComPtr")
    elif issubclass(type_, c_void_p):
        if isinstance(obj, Array):
            # struct doesn't support this conversion, though function does it.
            return cast(obj, c_void_p)
    elif issubclass(type_, _CFuncPtr):
        if isinstance(obj, type_):
            return obj
        elif callable(obj):
            r = type_(obj)
            easycast_keep_reference.append(r)
            return r
        elif obj is None:
            return type_()
    return obj


def get_type_hints(prototype, include_extras=False):
    if sys.version_info < (3, 10) and isinstance(prototype, type) and issubclass(prototype, (Structure, Union)):
        hints = _get_type_hints(prototype, localns=vars(prototype), include_extras=include_extras)
    else:
        hints = _get_type_hints(prototype, include_extras=include_extras)
    for name, type_ in hints.items():
        if not isinstance(type_, type):
            # generic
            pass
        elif type_ is type(None):
            hints[name] = None
        elif issubclass(type_, Enum):
            # Ctype's fundamental types are converted to python type transparently.
            # It is not applied to subclass.  Avoid it for Enum.
            hints[name] = type_.__bases__[1]
    return hints


def parse_arguments(funcname: str, params: list, args: tuple, kwargs: dict, variadic: bool) -> list:
    for k in params[: len(args)]:
        if k in kwargs:
            raise TypeError(f"{funcname}() got multiple values for argument '{k}'")
    for k in kwargs:
        if k not in params:
            raise TypeError(f"{funcname}() got an unexpected keyword argument '{k}'")
    nargs = len(args) + len(kwargs)
    if nargs < len(params):
        missing_count = len(params) - nargs
        missing_keys = ", ".join(f"'{k}'" for k in params[-missing_count:])
        raise TypeError(f"{funcname}() missing {missing_count} required positional arguments: {missing_keys}")
    if nargs > len(params) and not variadic:
        raise TypeError(f"{funcname}() takes {len(params)} positional arguments but {nargs} were given")
    return list(args) + list(kwargs[k] for k in params[len(args) :])


class Guid(Structure):
    _fields_ = [
        ("Data1", UInt32),
        ("Data2", UInt16),
        ("Data3", UInt16),
        ("Data4", Byte * 8),
    ]

    def __init__(self, val=None):
        if val is None:
            pass
        elif isinstance(val, self.__class__):
            self.Data1 = val.Data1
            self.Data2 = val.Data2
            self.Data3 = val.Data3
            self.Data4 = val.Data4
        elif isinstance(val, str):
            u = uuid.UUID(val)
            self.Data1 = u.time_low
            self.Data2 = u.time_mid
            self.Data3 = u.time_hi_version
            for i in range(8):
                self.Data4[i] = u.bytes[8 + i]
        elif isinstance(val, uuid.UUID):
            u = val
            self.Data1 = u.time_low
            self.Data2 = u.time_mid
            self.Data3 = u.time_hi_version
            for i in range(8):
                self.Data4[i] = u.bytes[8 + i]
        else:
            raise ValueError()

    def __str__(self):
        return f"{{{self.Data1:08x}-{self.Data2:04x}-{self.Data3:04x}-{self.Data4[0]:02x}{self.Data4[1]:02x}-{self.Data4[2]:02x}{self.Data4[3]:02x}{self.Data4[4]:02x}{self.Data4[5]:02x}{self.Data4[6]:02x}{self.Data4[7]:02x}}}"

    def __eq__(self, other):
        if not isinstance(other, Guid):
            raise ValueError(f"cannot compare with {type(other)}")
        return (
            self.Data1 == other.Data1
            and self.Data2 == other.Data2
            and self.Data3 == other.Data3
            and self.Data4[0] == other.Data4[0]
            and self.Data4[1] == other.Data4[1]
            and self.Data4[2] == other.Data4[2]
            and self.Data4[3] == other.Data4[3]
            and self.Data4[4] == other.Data4[4]
            and self.Data4[5] == other.Data4[5]
            and self.Data4[6] == other.Data4[6]
            and self.Data4[7] == other.Data4[7]
        )


def SUCCEEDED(hr):
    return hr >= 0


def FAILED(hr):
    return hr < 0


class ForeignFunction:
    def __init__(self, prototype, library, entry_point, variadic, dlltype, functype):
        self._prototype = prototype
        self._library = library
        self._entry_point = entry_point
        self._variadic = variadic
        self._dlltype = dlltype
        self._functype = functype
        self._delegate = None

    def _create_foreign_function_call(self):
        if self._entry_point is None:
            entry_point = self._prototype.__name__
        else:
            entry_point = self._entry_point
        if self._variadic:
            # Disable keyword argument for variadic function.
            params = None
        else:
            params = tuple((1, name) for name in self._prototype.__annotations__.keys() if name != "return")
        return ForeignFunctionCall(
            self._prototype, self._functype, [(entry_point, self._dlltype[self._library]), params], self._variadic
        )

    def __call__(self, *args, **kwargs):
        if self._delegate is None:
            self._delegate = self._create_foreign_function_call()
        return self._delegate(*args, **kwargs)


class ForeignFunctionCall:
    def __init__(self, prototype, functype, spec, variadic):
        hints = get_type_hints(prototype)
        self._restype = restype = hints.pop("return")
        if self._restype is c_char_p or self._restype is c_wchar_p:
            restype = UIntPtr
        self._prototype = prototype
        self._hints = hints
        self._variadic = variadic
        self._delegate = functype(restype, *self._hints.values())(*spec)

    def __call__(self, *args, **kwargs):
        _as_intptr = kwargs.pop("_as_intptr", False)
        cargs = self.make_args(args, kwargs)
        result = self._delegate(*cargs)
        return self.make_result(result, _as_intptr)

    def make_args(self, args, kwargs):
        pargs = parse_arguments(self._prototype.__qualname__, list(self._hints), args, kwargs, self._variadic)
        return [easycast(v, t) if t else v for v, t in zip_longest(pargs, self._hints.values())]

    def make_result(self, result, _as_intptr):
        if _as_intptr:
            if result is None:
                return 0
            return result
        if self._restype is c_char_p or self._restype is c_wchar_p:
            return self._restype(result).value
        return result


class ComMethod:
    def __init__(self, prototype, vtbl_index):
        self._prototype = prototype
        self._vtbl_index = vtbl_index
        self._delegate = None

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return types.MethodType(self.__call__, instance)

    def __call__(self, this, *args, **kwargs):
        if self._delegate is None:
            self._delegate = ComMethodCall(self._prototype, self._vtbl_index)
        return self._delegate(this, *args, **kwargs)


class ComMethodCall:
    def __init__(self, prototype, vtbl_index):
        hints = get_type_hints(prototype)
        self._restype = restype = hints.pop("return")
        if self._restype is c_char_p or self._restype is c_wchar_p:
            restype = UIntPtr
        params = tuple((1, name) for name in hints.keys())
        self._prototype = prototype
        self._hints = hints
        self._delegate = WINFUNCTYPE(restype, *self._hints.values())(vtbl_index, prototype.__name__, params)

    def __call__(self, this, *args, **kwargs):
        _as_intptr = kwargs.pop("_as_intptr", False)
        cargs = self.make_args(args, kwargs)
        result = self._delegate(this, *cargs)
        return self.make_result(result, _as_intptr)

    def make_args(self, args, kwargs):
        pargs = parse_arguments(self._prototype.__qualname__, list(self._hints), args, kwargs, False)
        return [easycast(v, t) for v, t in zip(pargs, self._hints.values())]  # >=3.10 strict=True

    def make_result(self, result, _as_intptr):
        if _as_intptr:
            if result is None:
                return 0
            return result
        if self._restype is c_char_p or self._restype is c_wchar_p:
            return self._restype(result).value
        return result


def cfunctype(library, entry_point=None, variadic=False):
    def decorator(prototype):
        return ForeignFunction(prototype, library, entry_point, variadic, cdll, CFUNCTYPE)

    return decorator


def winfunctype(library, entry_point=None):
    def decorator(prototype):
        return ForeignFunction(prototype, library, entry_point, False, windll, WINFUNCTYPE)

    return decorator


def commethod(vtbl_index):
    def decorator(prototype):
        return ComMethod(prototype, vtbl_index)

    return decorator


class ForeignFunctionPointer:
    def __init__(self, prototype, functype):
        self.__module__ = prototype.__module__  # referred by make_ready()
        self._prototype = prototype
        self._functype = functype

    def __commit__(self):
        hints = get_type_hints(self._prototype)

        if self._prototype.__name__ in sys.modules[self.__module__].__dict__:
            raise CircularReferenceResolved()

        restype = hints.pop("return")
        argtypes = list(hints.values())
        return self._functype(restype, *argtypes)


def cfunctype_pointer(prototype):
    return ForeignFunctionPointer(prototype, CFUNCTYPE)


def winfunctype_pointer(prototype):
    return ForeignFunctionPointer(prototype, WINFUNCTYPE)


class LazyLoader:
    def __init__(self, module):
        self._module = module
        self._lazyload = {}

    def register(self, name, prototype):
        self._lazyload[name] = prototype

    def setup(self):
        for name in self._lazyload:
            delattr(self._module, name)

    def __call__(self, name):
        try:
            prototype = self._lazyload[name]
        except KeyError:
            raise AttributeError(f"module '{self._module.__name__}' has no attribute '{name}'") from None
        try:
            setattr(self._module, name, prototype.__commit__())
            del self._lazyload[name]
        except CircularReferenceResolved:
            pass
        except AttributeError as e:
            # AttributeError will be ignored in __getattr__().
            raise RuntimeError("Unexpected error") from e
        return getattr(self._module, name)


class CircularReferenceResolved(Exception):
    pass


class ConstantLazyLoader:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self.__annotations__ = {}

    def __set_name__(self, owner, name):
        # get_type_hints() referes __globals__.
        self.__globals__ = owner.__dict__
        self.__annotations__["self"] = owner.__annotations__[name]

    def __commit__(self):
        cls = get_type_hints(self)["self"]
        return cls(*self._args, **self._kwargs)


class UnicodeAlias:
    def __init__(self, unicode_name):
        self._unicode_name = unicode_name
        self._owner = None

    def __set_name__(self, owner, name):
        self._owner = owner

    def __commit__(self):
        return getattr(self._owner, self._unicode_name)


def make_ready(module_name: str) -> None:
    module = sys.modules[module_name]

    module.__getattr__ = lazy_loader = LazyLoader(module)

    for name, value in vars(module).items():
        if isinstance(value, ConstantLazyLoader):
            value.__set_name__(module, name)
            lazy_loader.register(name, value)
        elif isinstance(value, UnicodeAlias):
            value.__set_name__(module, name)
            lazy_loader.register(name, value)
        elif hasattr(value, "__commit__") and value.__module__ == module_name:
            lazy_loader.register(name, value)

    lazy_loader.setup()



from .win32 import *

# Initialize COM Multithreaded Apartment.
# Call CoInitializeEx(None, COINIT_APARTMENTTHREADED) explicitly for Single-Threaded Apartment.
cookie = VoidPtr()
hr = windll["ole32.dll"].CoIncrementMTAUsage(pointer(cookie))
if FAILED(hr):
    raise WinError(hr)
del hr
del cookie



from __future__ import annotations

import asyncio
import inspect
import logging
import sys
import types
import uuid
from collections.abc import Iterable, MutableSequence, Sequence
from contextlib import ExitStack
from ctypes import (
    POINTER,
    Structure,
    _SimpleCData,
    addressof,
    c_void_p,
    cast,
    pointer,
    sizeof,
    wstring_at,
)
from functools import partial
from typing import Any, Generic, TypeVar, _GenericAlias

if sys.version_info < (3, 9):
    from typing_extensions import Annotated, Tuple, get_args, get_origin  # noqa: F401
else:
    from typing import Annotated, Tuple, get_args, get_origin  # noqa: F401

from win32more import (
    FAILED,
    WINFUNCTYPE,
    Boolean,
    Byte,
    Char,
    ComMethod,
    ComPtr,
    Double,
    Enum,
    Guid,
    Int32,
    Int64,
    Single,
    String,
    UInt32,
    UInt64,
    Void,
    VoidPtr,
    WinError,
    asyncui,
    easycast,
    get_type_hints,
    parse_arguments,
    windll,
)
from win32more.Windows.Win32.Foundation import (
    BSTR,
    E_FAIL,
    E_NOINTERFACE,
    HRESULT,
    S_OK,
    SysAllocString,
    SysFreeString,
    SysStringLen,
)
from win32more.Windows.Win32.System.Com import (
    CoTaskMemAlloc,
    CoTaskMemFree,
    GetErrorInfo,
    IAgileObject,
    IErrorInfo,
    IUnknown,
    SetErrorInfo,
)
from win32more.Windows.Win32.System.Ole import CreateErrorInfo, ICreateErrorInfo
from win32more.Windows.Win32.System.WinRT import (
    HSTRING,
    PFNGETACTIVATIONFACTORY,
    BaseTrust,
    IActivationFactory,
    IInspectable,
    IRestrictedErrorInfo,
    RoGetActivationFactory,
    RoOriginateError,
    TrustLevel,
    WindowsCreateString,
    WindowsDeleteString,
    WindowsGetStringRawBuffer,
)

K = TypeVar("K")
T = TypeVar("T")
V = TypeVar("V")
TProgress = TypeVar("TProgress")
TResult = TypeVar("TResult")
TSender = TypeVar("TSender")

logger = logging.getLogger(__name__)


class event:
    def __init__(self):
        self._name = None

    def __set_name__(self, owner, name):
        self._name = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return event_setter(instance, self._name)


class event_setter:
    def __init__(self, instance, name):
        self._instance = instance
        self._name = name

    def __iadd__(self, callback):
        getattr(self._instance, f"add_{self._name}")(callback)


def ComPtr_commit(cls):
    cls._hints_ = get_type_hints(cls)
    if cls._hints_["extends"] is None:
        return cls
    # Generic class have multiple base class (Generic[], ComPtr).
    bases = []
    for type_ in cls.__bases__:
        if type_ is ComPtr:
            type_ = cls._hints_["extends"]
        bases.append(type_)
    if "implements" in cls._hints_:
        bases.extend(get_args(cls._hints_["implements"]))
    cls.__bases__ = tuple(bases)
    if "__orig_bases__" in cls.__dict__:
        orig_bases = []
        for type_ in cls.__bases__:
            if type_ is ComPtr:
                type_ = cls._hints_["extends"]
            orig_bases.append(type_)
        cls.__orig_bases__ = tuple(orig_bases)
    return cls


def ComPtr_as(self, cls):
    if cls is str:
        return unbox_value(self)
    elif is_generic_alias(cls):
        iid = _ro_get_parameterized_type_instance_iid(cls)
    elif "_iid_" in cls.__dict__:
        iid = cls._iid_
    elif "default_interface" in cls._hints_:
        iid = cls._hints_["default_interface"]._iid_
    else:
        raise RuntimeError("no _iid_ found")
    instance = cls(own=True)
    hr = self.QueryInterface(pointer(iid), pointer(instance))
    if FAILED(hr):
        raise WinError(hr)
    return instance


ComPtr.__commit__ = classmethod(ComPtr_commit)
ComPtr.as_ = ComPtr_as


def winrt_easycast(obj, type_):
    from win32more.Windows.Foundation import IReference
    from win32more.Windows.Foundation.Collections import IVector
    from win32more.winrt.vector import Vector

    if type_ is IInspectable:
        if isinstance(obj, str):
            return box_value(obj)
    elif issubclass(_get_origin_or_itself(type_), IVector):
        if isinstance(obj, list):
            return Vector[get_args(type_)[0]](obj)
    elif issubclass(_get_origin_or_itself(type_), IReference):
        # FIXME: Should I check obj is T of IReference[T]?
        if obj is None:
            return IReference(None)
        return box_value(obj).as_(type_)
    elif is_generic_alias(type_):
        # Do not propagate generic type
        return obj
    return easycast(obj, type_)


# FIXME: Add more conversion.
def box_value(value: object) -> IInspectable:
    from win32more.Windows.Foundation import PropertyValue

    if isinstance(value, str):
        return PropertyValue.CreateString(value)
    elif isinstance(value, bool):
        return PropertyValue.CreateBoolean(value)
    raise NotImplementedError(f"box_value: {type(value)}")


def unbox_value(value: IInspectable):
    from win32more.Windows.Foundation import IPropertyValue, PropertyType

    property_value = value.as_(IPropertyValue)

    if property_value.Type == PropertyType.Empty:
        return None
    elif property_value.Type == PropertyType.UInt8:
        return property_value.GetUInt8()
    elif property_value.Type == PropertyType.Int16:
        return property_value.GetInt16()
    elif property_value.Type == PropertyType.UInt16:
        return property_value.GetUInt16()
    elif property_value.Type == PropertyType.Int32:
        return property_value.GetInt32()
    elif property_value.Type == PropertyType.UInt32:
        return property_value.GetUInt32()
    elif property_value.Type == PropertyType.Int64:
        return property_value.GetInt64()
    elif property_value.Type == PropertyType.UInt64:
        return property_value.GetUInt64()
    elif property_value.Type == PropertyType.Single:
        return property_value.GetSingle()
    elif property_value.Type == PropertyType.Double:
        return property_value.GetDouble()
    elif property_value.Type == PropertyType.Char16:
        return property_value.GetChar16()
    elif property_value.Type == PropertyType.Boolean:
        return property_value.GetBoolean()
    elif property_value.Type == PropertyType.String:
        return property_value.GetString()
    elif property_value.Type == PropertyType.Inspectable:
        return value
    elif property_value.Type == PropertyType.DateTime:
        return property_value.GetDateTime()
    elif property_value.Type == PropertyType.TimeSpan:
        return property_value.GetTimeSpan()
    elif property_value.Type == PropertyType.Guid:
        return property_value.GetGuid()
    elif property_value.Type == PropertyType.Point:
        return property_value.GetPoint()
    elif property_value.Type == PropertyType.Size:
        return property_value.GetSize()
    elif property_value.Type == PropertyType.Rect:
        return property_value.GetRect()
    elif property_value.Type == PropertyType.UInt8Array:
        r = []
        property_value.GetUInt8Array(r)
        return r
    elif property_value.Type == PropertyType.Int16Array:
        r = []
        property_value.GetInt16Array(r)
        return r
    elif property_value.Type == PropertyType.UInt16Array:
        r = []
        property_value.GetUInt16Array(r)
        return r
    elif property_value.Type == PropertyType.Int32Array:
        r = []
        property_value.GetInt32Array(r)
        return r
    elif property_value.Type == PropertyType.UInt32Array:
        r = []
        property_value.GetUInt32Array(r)
        return r
    elif property_value.Type == PropertyType.Int64Array:
        r = []
        property_value.GetInt64Array(r)
        return r
    elif property_value.Type == PropertyType.UInt64Array:
        r = []
        property_value.GetUInt64Array(r)
        return r
    elif property_value.Type == PropertyType.SingleArray:
        r = []
        property_value.GetSingleArray(r)
        return r
    elif property_value.Type == PropertyType.DoubleArray:
        r = []
        property_value.GetDoubleArray(r)
        return r
    elif property_value.Type == PropertyType.Char16Array:
        r = []
        property_value.GetChar16Array(r)
        return r
    elif property_value.Type == PropertyType.BooleanArray:
        r = []
        property_value.GetBooleanArray(r)
        return r
    elif property_value.Type == PropertyType.StringArray:
        r = []
        property_value.GetStringArray(r)
        return r
    elif property_value.Type == PropertyType.InspectableArray:
        r = []
        property_value.GetInspectableArray(r)
        return r
    elif property_value.Type == PropertyType.DateTimeArray:
        r = []
        property_value.GetDateTimeArray(r)
        return r
    elif property_value.Type == PropertyType.TimeSpanArray:
        r = []
        property_value.GetTimeSpanArray(r)
        return r
    elif property_value.Type == PropertyType.GuidArray:
        r = []
        property_value.GetGuidArray(r)
        return r
    elif property_value.Type == PropertyType.PointArray:
        r = []
        property_value.GetPointArray(r)
        return r
    elif property_value.Type == PropertyType.SizeArray:
        r = []
        property_value.GetSizeArray(r)
        return r
    elif property_value.Type == PropertyType.RectArray:
        r = []
        property_value.GetRectArray(r)
        return r
    # elif property_value.Type == PropertyType.OtherType:
    # elif property_value.Type == PropertyType.OtherTypeArray:
    else:
        raise NotImplementedError(f"unbox_value: {property_value.Type}")


def generic_get_type_hints(prototype, cls):
    hints = get_type_hints(prototype)
    if is_generic_alias(cls):
        gs = GenericSpecializer.from_generic_alias(cls)
        hints = {key: gs.specialize_parameter(parameter) for key, parameter in hints.items()}
    return hints


class GenericSpecializer:
    def __init__(self, parameter_to_type_map: dict[TypeVar, type]) -> None:
        self._parameter_to_type_map = parameter_to_type_map

    @classmethod
    def from_generic_alias(cls, specialized_generic_alias: _GenericAlias) -> GenericSpecializer:
        parameters = get_origin(specialized_generic_alias).__parameters__
        args = get_args(specialized_generic_alias)
        return GenericSpecializer(dict(zip(parameters, args)))

    def specialize_generic_alias(self, parameterized_generic_alias: _GenericAlias) -> _GenericAlias:
        parameters = get_args(parameterized_generic_alias)
        args = self.specialize_parameters(parameters)
        return get_origin(parameterized_generic_alias)[tuple(args)]

    def specialize_parameters(self, parameters: list[_GenericAlias | TypeVar | type]) -> list[_GenericAlias | type]:
        return [self.specialize_parameter(parameter) for parameter in parameters]

    def specialize_parameter(self, parameter: _GenericAlias | TypeVar | type) -> _GenericAlias | type:
        if is_generic_alias(parameter):
            return self.specialize_generic_alias(parameter)
        elif isinstance(parameter, TypeVar):
            return self._parameter_to_type_map[parameter]
        else:
            return parameter


# types.get_original_bases
def _get_original_bases(cls):
    return cls.__dict__.get("__orig_bases__", cls.__bases__)


def _get_original_class(instance):
    return instance.__dict__.get("__orig_class__", instance.__class__)


def _get_origin_or_itself(cls):
    return get_origin(cls) or cls


def _get_specialized_bases(cls):
    if not is_generic_alias(cls):
        return _get_original_bases(cls)
    gs = GenericSpecializer.from_generic_alias(cls)
    bases = []
    for base in _get_original_bases(_get_origin_or_itself(cls)):
        if get_origin(base) is Generic:
            continue
        bases.append(gs.specialize_parameter(base))
    return bases


# Dummy type for list[T]
class PassArray(Generic[T]):
    def __init__(self, type_, seq, cargs):
        self._type = type_

        if isinstance(seq, Sequence):
            if self._type is WinRT_String:
                self._seq = [WinRT_String(s, own=True) for s in seq]
            else:
                self._seq = seq
            cargs.append(len(self._seq))
            cargs.append((self._type * len(self._seq))(*self._seq))
        else:
            raise TypeError(f"cannot convert {type(seq)} to PassArray[{type_.__name__}]")

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


class PassArrayCallback:
    def __init__(self, type_, size, ptr, pyargs):
        self._type = type_
        self._size = size
        self._ptr = ptr
        self._lst = []
        for i in range(size):
            if type_ is WinRT_String:
                self._lst.append(_windows_get_string_raw_buffer(ptr[i]))
            elif not is_simple_cdata(type_):
                self._lst.append(type_.from_buffer_copy(ptr[i]))
            else:
                self._lst.append(ptr[i])
        pyargs.append(self._lst)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


# Dummy type for list[T]
class FillArray(Generic[T]):
    def __init__(self, type_, seq, cargs):
        self._type = type_

        if isinstance(seq, MutableSequence):
            self._seq = seq
            self._ptr = (type_ * len(seq))()
            cargs.append(len(self._seq))
            cargs.append(self._ptr)
        else:
            raise TypeError(f"cannot convert {type(seq)} to FillArray[{type_.__name__}]")

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            return

        if self._type is WinRT_String:
            self._seq[:] = [s.strvalue for s in self._ptr[:]]
            for s in self._ptr[:]:
                s.clear()
        elif is_com_class(self._type):
            self._seq[:] = self._ptr[:]
            for p in self.seq:
                p._own = True
        else:
            self._seq[:] = self._ptr[:]


class FillArrayCallback:
    def __init__(self, type_, size, ptr, pyargs):
        self._type = type_
        self._size = size
        self._ptr = ptr
        self._lst = [None] * size
        pyargs.append(self._lst)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            return

        for i, v in enumerate(self._lst):
            if self._type is WinRT_String:
                self._ptr[i] = WinRT_String(v)
            elif is_com_class(self._type):
                if v:
                    v.AddRef()
                self._ptr[i] = v
            else:
                self._ptr[i] = v


# Dummy type for list[T]
class ReceiveArray(Generic[T]):
    def __init__(self, type_, seq, cargs):
        self._type = type_

        if isinstance(seq, MutableSequence):
            self._seq = seq
            self._size = UInt32()
            self._ptr = POINTER(type_)()
            cargs.append(pointer(self._size))
            cargs.append(pointer(self._ptr))
        else:
            raise TypeError(f"cannot convert {type(seq)} to ReceiveArray[{type_.__name__}]")

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            return

        if self._type is WinRT_String:
            self._seq[:] = [s.strvalue for s in self._ptr[: self._size.value]]
            for s in self._ptr[: self._size.value]:
                s.clear()
        elif is_com_class(self._type):
            self._seq[:] = [self._type.from_buffer_copy(p) for p in self._ptr[: self._size.value]]
            for p in self._seq:
                p._own = True
        elif not is_simple_cdata(self._type):
            self._seq[:] = [self._type.from_buffer_copy(p) for p in self._ptr[: self._size.value]]
        else:
            self._seq[:] = self._ptr[: self._size.value]
        CoTaskMemFree(self._ptr)


class ReceiveArrayCallback:
    def __init__(self, type_, psize, pptr, pyargs):
        self._type = type_
        self._psize = psize
        self._pptr = pptr
        self._lst = []
        pyargs.append(self._lst)

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            return
        self.handle_result(self._type, self._psize, self._pptr, self._lst)

    @staticmethod
    def handle_result(type_, psize, pptr, seq):
        if not isinstance(seq, Sequence):
            raise ValueError(f"Sequence is expected: {seq}")

        psize[0] = len(seq)
        ptr = CoTaskMemAlloc(len(seq) * sizeof(type_))
        if not ptr:
            raise WinError()
        pptr[0] = cast(ptr, POINTER(type_))
        for i, v in enumerate(seq):
            if type_ is WinRT_String:
                pptr[0][i] = WinRT_String(v)
            elif is_com_class(type_):
                if v:
                    v.AddRef()
                pptr[0][i] = v
            else:
                pptr[0][i] = v


# FIXME: Not work for array and struct entry.
class WinRT_String(HSTRING):
    def __init__(self, obj=None, own=False):
        if obj is None:
            hs = HSTRING(0)
        elif isinstance(obj, HSTRING):
            hs = obj
        elif isinstance(obj, String):
            hs = _windows_create_string(obj.value)
        elif isinstance(obj, str):
            hs = _windows_create_string(obj)
        else:
            raise ValueError(obj)
        # https://github.com/python/cpython/issues/73456
        # super().__init__(hs.value)
        self.value = hs.value
        self._own = own

    def __del__(self):
        if self and getattr(self, "_own", False):
            self.clear()

    def clear(self):
        if self.value:
            hr = WindowsDeleteString(self)
            if FAILED(hr):
                raise WinError(hr)
            self.value = 0

    @classmethod
    def from_param(cls, obj):
        if isinstance(obj, String):
            return cls(obj, own=True)
        elif isinstance(obj, str):
            return cls(obj, own=True)
        else:
            return cls(obj, own=False)

    @property
    def strvalue(self):
        return _windows_get_string_raw_buffer(self)


class WinrtMethod:
    def __init__(self, prototype, vtbl_index):
        self._prototype = prototype
        self._vtbl_index = vtbl_index
        self._generic_delegate = {}

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return types.MethodType(self.__call__, instance)

    def __call__(self, this, *args, **kwargs):
        cls = _get_original_class(this)
        generic_args = get_args(cls)
        if generic_args not in self._generic_delegate:
            self._generic_delegate[generic_args] = WinrtMethodCall(self._prototype, self._vtbl_index, cls)
        return self._generic_delegate[generic_args](this, *args, **kwargs)


class WinrtMethodCall:
    def __init__(self, prototype, vtbl_index, cls):
        hints = generic_get_type_hints(prototype, cls)
        restype = hints.pop("return")
        argtypes = []
        params = []
        for name, type_ in hints.items():
            self._add_parameter(argtypes, params, name, type_)
        self._add_restype(argtypes, params, restype)
        self.delegate = WINFUNCTYPE(HRESULT, *argtypes)(vtbl_index, prototype.__name__, tuple(params))
        self.restype = restype
        self._prototype = prototype
        self.hints = hints

    def __call__(self, this, *args, **kwargs):
        with ExitStack() as exitstack:
            cargs = []
            pargs = parse_arguments(self._prototype.__qualname__, list(self.hints), args, kwargs, False)
            for value, type_ in zip(pargs, self.hints.values()):  # >=3.10 strict=True
                self._add_argument(cargs, value, type_, exitstack)
            result = self._add_result(cargs, self.restype, exitstack)
            hr = self.delegate(this, *cargs)
            if FAILED(hr):
                # FIXME: Is restricted error obtained here always associated with this hr?
                raise ComError(hr)
        return self._handle_result(result)

    def _add_parameter(self, argtypes: list[type], params: list[tuple[int, str]], name: str, type_: type) -> None:
        if is_passarray_class(type_):
            argtypes.append(UInt32)
            argtypes.append(POINTER(get_args(type_)[0]))
            params.append((1, f"{name}_length"))
            params.append((1, name))
        elif is_fillarray_class(type_):
            argtypes.append(UInt32)
            argtypes.append(POINTER(get_args(type_)[0]))
            params.append((1, f"{name}_length"))
            params.append((1, name))
        elif is_receivearray_class(type_):
            argtypes.append(POINTER(UInt32))
            argtypes.append(POINTER(POINTER(get_args(type_)[0])))
            params.append((1, f"{name}_length"))
            params.append((1, name))
        elif is_delegate_class(type_):
            argtypes.append(MulticastDelegateImpl)
            params.append((1, name))
        elif is_generic_alias(type_):
            argtypes.append(get_origin(type_))
            params.append((1, name))
        else:
            argtypes.append(type_)
            params.append((1, name))

    def _add_restype(self, argtypes: list[type], params: list[tuple[int, str]], restype: type) -> None:
        if restype is Void:
            pass
        elif is_receivearray_class(restype):
            argtypes.append(POINTER(UInt32))
            argtypes.append(POINTER(POINTER(get_args(restype)[0])))
            params.append((1, "return_length"))
            params.append((1, "return"))
        elif is_generic_alias(restype):
            argtypes.append(POINTER(get_origin(restype)))
            params.append((1, "return"))
        else:
            argtypes.append(POINTER(restype))
            params.append((1, "return"))

    def _add_argument(self, cargs: list[Any], value: Any, type_: type, exitstack: ExitStack) -> None:
        if is_passarray_class(type_):
            exitstack.enter_context(PassArray(get_args(type_)[0], value, cargs))
        elif is_fillarray_class(type_):
            exitstack.enter_context(FillArray(get_args(type_)[0], value, cargs))
        elif is_receivearray_class(type_):
            exitstack.enter_context(ReceiveArray(get_args(type_)[0], value, cargs))
        elif is_delegate_class(type_):
            cargs.append(MulticastDelegateImpl(type_, value))
        elif is_com_instance(value) and is_com_class(type_):
            cargs.append(value.as_(type_))
        else:
            cargs.append(winrt_easycast(value, type_))

    def _add_result(self, cargs: list[Any], restype: type, exitstack) -> None:
        if self.restype is Void:
            result = None
        elif is_receivearray_class(self.restype):
            result = []
            exitstack.enter_context(ReceiveArray(get_args(self.restype)[0], result, cargs))
        elif is_com_class(self.restype):
            result = self.restype(own=True)
            cargs.append(pointer(result))
        elif self.restype is WinRT_String:
            result = self.restype(own=True)
            cargs.append(pointer(result))
        else:
            result = self.restype()
            cargs.append(pointer(result))
        return result

    def _handle_result(self, result: Any) -> Any:
        from win32more.Windows.Foundation import IReference

        if self.restype is Void:
            return None
        elif issubclass(_get_origin_or_itself(self.restype), IReference):
            if not result:
                return None
            return unbox_value(result)
        elif issubclass(_get_origin_or_itself(self.restype), Enum):
            return result.value
        elif is_com_class(self.restype):
            if not result:
                return None
            return result
        elif self.restype is WinRT_String:
            return result.strvalue
        elif is_simple_cdata(_get_origin_or_itself(self.restype)):
            return result.value
        return result


def winrt_commethod(vtbl_index):
    def decorator(prototype):
        return WinrtMethod(prototype, vtbl_index)

    return decorator


class winrt_mixinmethod:
    def __init__(self, prototype):
        self._prototype = prototype

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return types.MethodType(self.__call__, instance)

    def __call__(self, instance, *args, **kwargs):
        hints = get_type_hints(self._prototype)
        interface_instance = instance.as_(hints["self"])
        return getattr(interface_instance, self._prototype.__name__)(*args, **kwargs)

    def argcount(self):
        return len(self._prototype.__annotations__) - 2


# Cls[T]?
def is_generic_alias(cls):
    return isinstance(cls, _GenericAlias)


# Cls[T]()?
def is_generic_instance(obj):
    return isinstance(obj, Generic)


def is_delegate_class(cls):
    return issubclass(_get_origin_or_itself(cls), MulticastDelegate)


def is_com_class(cls):
    return issubclass(_get_origin_or_itself(cls), ComPtr)


def is_com_instance(obj):
    return isinstance(obj, ComPtr)


def is_passarray_class(cls):
    return issubclass(_get_origin_or_itself(cls), PassArray)


def is_fillarray_class(cls):
    return issubclass(_get_origin_or_itself(cls), FillArray)


def is_receivearray_class(cls):
    return issubclass(_get_origin_or_itself(cls), ReceiveArray)


# check if ptr[0] returns python primitive?
def is_simple_cdata(cls):
    return cls.__base__ is _SimpleCData


class winrt_classmethod:
    def __init__(self, prototype):
        self._prototype = prototype

    def __get__(self, instance, owner=None):
        return types.MethodType(self.__call__, owner)

    def __call__(self, cls, *args, **kwargs):
        hints = get_type_hints(self._prototype)
        factory_class = hints["cls"]
        factory = _ro_get_activation_factory(cls._classid_).as_(factory_class)
        return getattr(factory, self._prototype.__name__)(*args, **kwargs)

    def argcount(self):
        return len(self._prototype.__annotations__) - 2


def winrt_factorymethod(prototype):
    return winrt_classmethod(prototype)


class winrt_activatemethod:
    def __init__(self, prototype):
        self._prototype = prototype

    def __get__(self, instance, owner=None):
        return types.MethodType(self.__call__, owner)

    def __call__(self, cls):
        return _ro_activate_instance(cls._classid_).as_(cls)

    def argcount(self):
        return 0


class winrt_overload:
    def __init__(self, func: winrt_mixinmethod | winrt_classmethod | winrt_activatemethod) -> None:
        self.funcs = [func]

    def register(self, func: winrt_mixinmethod | winrt_classmethod | winrt_activatemethod):
        self.funcs.append(func)
        return self

    def __get__(self, instance, owner=None):
        if instance is None:
            return types.MethodType(self._call_classmethod, owner)
        else:
            return types.MethodType(self._call_method, instance)

    def _call_classmethod(self, cls, *args, **kwargs):
        for func in self.funcs:
            if isinstance(func, (winrt_classmethod, winrt_activatemethod)) and len(args) == func.argcount():
                return func(cls, *args, **kwargs)
        raise ValueError("no matched method")

    def _call_method(self, instance, *args, **kwargs):
        for func in self.funcs:
            if isinstance(func, winrt_mixinmethod) and len(args) == func.argcount():
                return func(instance, *args, **kwargs)
        raise ValueError("no matched method")


def _windows_create_string(s: str) -> HSTRING:
    hs = HSTRING()
    hr = WindowsCreateString(s, len(s), hs)
    if FAILED(hr):
        raise WinError(hr)
    return hs


def _windows_get_string_raw_buffer(hs: HSTRING) -> str:
    length = UInt32()
    bufaddr = WindowsGetStringRawBuffer(hs, length, _as_intptr=True)
    return wstring_at(bufaddr, length.value)


def _ro_get_activation_factory(classid: str) -> IActivationFactory:
    hs = _windows_create_string(classid)
    factory = IActivationFactory(own=True)
    hr = RoGetActivationFactory(hs, IActivationFactory._iid_, factory)
    WindowsDeleteString(hs)
    if FAILED(hr):
        return _get_runtime_activation_factory(classid)
    return factory


def _ro_activate_instance(classid: str) -> IInspectable:
    factory = _ro_get_activation_factory(classid)
    instance = IInspectable(own=True)
    hr = factory.ActivateInstance(instance)
    if FAILED(hr):
        raise WinError(hr)
    return instance


# cppwinrt: base.h
def _get_runtime_activation_factory(classid: str) -> IActivationFactory:
    name = classid
    while "." in name:
        name, _ = name.rsplit(".", 1)
        dllname = name + ".dll"

        try:
            DllGetActivationFactory = cast(windll[dllname].DllGetActivationFactory, PFNGETACTIVATIONFACTORY)
        except AttributeError:
            continue

        hs = _windows_create_string(classid)
        factory = IActivationFactory(own=True)
        hr = DllGetActivationFactory(hs, factory)
        WindowsDeleteString(hs)
        if FAILED(hr):
            continue
        return factory

    raise WinError(-2147024770)  # HRESULT_FROM_WIN32(ERROR_MOD_NOT_FOUND)


# https://learn.microsoft.com/en-us/uwp/winrt-cref/winrt-type-system
#
# signature_octets => guid_to_octets(wrt_pinterface_namespace) string_to_utf8_octets(ptype_instance_signature)
#         wrt_pinterface_namespace => "11f47ad5-7b73-42c0-abae-878b1e16adee"
#         ptype_instance_signature => pinterface_instance_signature | pdelegate_instance_ signature
#         pinterface_instance_signature => "pinterface(" piid_guid  ";" args ")"
#         pdelegate_instance_signature => "pinterface(" piid_guid ";" args ")"
#         piid_guid => guid
#         args => arg | arg ";" args
#         arg => type_signature
#         type_signature => base_type_identifer | com_interface_signature | interface_signature | delegate_signature  | interface_group_signature | runtime_class_signature | struct_signature | enum_signature | pinterface_instance_signature | pdelegate_instance_signature
#         com_interface_signature => "cinterface(IInspectable)"
#         base_type_identifier is defined below
#         interface_signature => guid
#         interface_group_signature => "ig(" interface_group_name ";" default_interface ")"
#         runtime_class_signature => "rc(" runtime_class_name ";" default_interface ")"
#         default_interface => type_signature
#         struct_signature => "struct(" struct_name ";" args ")"
#         enum_signature => "enum(" enum_name ";" enum_underlying_type ")"
#         enum_underlying_type => type_signature
#         delegate_signature => "delegate(" guid ")"
#         guid => "{" dashed_hex "}"
#         dashed_hex is the format that uuidgen writes in when passed no arguments.
#             dashed_hex => hex{8} "-" hex{4} "-" hex{4} "-" hex{4} "-" hex{12}
#             hex => [0-9a-f]
def _ro_get_parameterized_type_instance_iid(ga: _GenericAlias) -> Guid:
    wrt_pinterface_namespace = uuid.UUID("{11f47ad5-7b73-42c0-abae-878b1e16adee}")
    ptype_instance_signature = _get_type_signature(ga)
    return Guid(uuid.uuid5(wrt_pinterface_namespace, ptype_instance_signature))


# FIXME: not completed
def _get_type_signature(cls) -> str:
    if cls is IInspectable:
        return "cinterface(IInspectable)"
    elif isinstance(cls, _GenericAlias):
        piid_guid = str(cls._piid_)
        args = ";".join(_get_type_signature(arg) for arg in get_args(cls))
        return f"pinterface({piid_guid};{args})"
    elif issubclass(cls, ComPtr) and "_iid_" in cls.__dict__:
        return str(cls._iid_)
    elif issubclass(cls, ComPtr):
        default_interface = cls._hints_["default_interface"]
        args = _get_type_signature(default_interface)
        return f"rc({cls._classid_};{args})"
    elif issubclass(cls, WinRT_String):
        return "string"
    elif issubclass(cls, Char):
        return "c2"
    elif issubclass(cls, Int32):
        return "i4"
    elif issubclass(cls, Int64):
        return "i8"
    elif issubclass(cls, Byte):
        return "u1"
    elif issubclass(cls, UInt32):
        return "u4"
    elif issubclass(cls, UInt64):
        return "u8"
    elif issubclass(cls, Single):
        return "f4"
    elif issubclass(cls, Double):
        return "f8"
    elif issubclass(cls, Boolean):
        return "b1"
    elif issubclass(cls, Guid):
        return "g16"
    else:
        raise NotImplementedError()


class Vtbl(Structure):
    _fields_ = [("lpvtbl", POINTER(c_void_p))]

    def __init__(self, owner, interface):
        self._owner = owner
        self._interface = interface
        self._keep_reference_in_python_world_ = []
        self.lpvtbl = self._make_lpvtbl()

    def _make_lpvtbl(self):
        methods = self._interface_methods()
        vtbl_size = max(method._vtbl_index for method in methods) + 1
        lpvtbl = (c_void_p * vtbl_size)()
        for method in methods:
            if isinstance(method, WinrtMethod):
                lpvtbl[method._vtbl_index] = self._make_thunk_winrt(method)
            else:
                lpvtbl[method._vtbl_index] = self._make_thunk_com(method)
        return lpvtbl

    def _interface_methods(self):
        if sys.version_info < (3, 10) and is_generic_alias(self._interface):
            interface = get_origin(self._interface)
        else:
            interface = self._interface
        methods = []
        for name in dir(interface):
            if isinstance(getattr(interface, name), (ComMethod, WinrtMethod)):
                methods.append(getattr(interface, name))
        return methods

    def _make_thunk_com(self, method):
        hints = generic_get_type_hints(method._prototype, self._interface)
        restype = hints.pop("return")
        argtypes = [self._make_allocator(t) for t in hints.values()]
        closure = partial(self._com_callback_error_check, getattr(self._owner, method._prototype.__name__))
        thunk = WINFUNCTYPE(restype, c_void_p, *argtypes)(closure)
        self._keep_reference_in_python_world_.append(thunk)
        return cast(thunk, c_void_p)

    def _make_thunk_winrt(self, method):
        hints = generic_get_type_hints(method._prototype, self._interface)
        restype = hints.pop("return")
        argtypes = []
        for type_ in hints.values():
            if is_passarray_class(type_):
                argtypes.append(UInt32)
                argtypes.append(POINTER(get_args(type_)[0]))
            elif is_fillarray_class(type_):
                argtypes.append(UInt32)
                argtypes.append(POINTER(get_args(type_)[0]))
            elif is_receivearray_class(type_):
                argtypes.append(POINTER(UInt32))
                argtypes.append(POINTER(POINTER(get_args(type_)[0])))
            else:
                argtypes.append(self._make_allocator(type_))
        if restype is Void:
            pass
        elif is_receivearray_class(restype):
            argtypes.append(POINTER(UInt32))
            argtypes.append(POINTER(POINTER(get_args(restype)[0])))
        elif is_generic_alias(restype):
            argtypes.append(POINTER(get_origin(restype)))
        else:
            argtypes.append(POINTER(restype))
        closure = partial(
            self._winrt_callback_error_check, getattr(self._owner, method._prototype.__name__), restype, hints
        )
        thunk = WINFUNCTYPE(HRESULT, c_void_p, *argtypes)(closure)
        self._keep_reference_in_python_world_.append(thunk)
        return cast(thunk, c_void_p)

    # allocate winrt runtime class without constructor.
    def _make_allocator(self, t):
        if is_generic_alias(t) or issubclass(t, ComPtr):
            return type(c_void_p)("Allocator", (c_void_p,), {"__new__": lambda _: t(own=False)})
        return t

    def _com_callback_error_check(self, callback, this, *args):
        try:
            return callback(*args)
        except Exception as exc:
            logger.exception(f"Unhandled exception caught: {callback}{args}")
            self._set_error_info(exc)
            return E_FAIL

    def _set_error_info(self, exc):
        info = ICreateErrorInfo(own=True)
        hr = CreateErrorInfo(info)
        if FAILED(hr):
            return
        description = SysAllocString(f"{exc.__class__.__name__}: {exc}", _as_intptr=True)
        if not description:
            return
        hr = info.SetDescription(description)
        if FAILED(hr):
            return
        SetErrorInfo(0, info.as_(IErrorInfo))

    def _winrt_callback_error_check(self, callback, restype, hints, this, *args):
        try:
            self._winrt_callback(callback, restype, hints, this, args)
            return S_OK
        except Exception as exc:
            logger.exception(f"Unhandled exception caught: {callback}{args}")
            self._originate_error(exc)
            return E_FAIL

    def _originate_error(self, exc):
        try:
            msg = _windows_create_string(f"{exc.__class__.__name__}: {exc}")
        except OSError:
            return
        RoOriginateError(E_FAIL, msg)

    def _winrt_callback(self, callback, restype, hints, this, args):
        args = list(args)
        with ExitStack() as exitstack:
            pyargs = []
            for type_ in hints.values():
                self._add_argument(type_, args, pyargs, exitstack)
            result = callback(*pyargs)
        self._handle_result(restype, args, result)

    def _add_argument(self, type_: type, args: list[Any], pyargs: list[Any], exitstack) -> None:
        if is_passarray_class(type_):
            exitstack.enter_context(PassArrayCallback(get_args(type_)[0], args.pop(0), args.pop(0), pyargs))
        elif is_fillarray_class(type_):
            exitstack.enter_context(FillArrayCallback(get_args(type_)[0], args.pop(0), args.pop(0), pyargs))
        elif is_receivearray_class(type_):
            exitstack.enter_context(ReceiveArrayCallback(get_args(type_)[0], args.pop(0), args.pop(0), pyargs))
        elif type_ is WinRT_String:
            pyargs.append(_windows_get_string_raw_buffer(args.pop(0)))
        else:
            pyargs.append(args.pop(0))

    def _handle_result(self, restype: type, args: list[Any], result: Any) -> Any:
        if restype is Void:
            pass
        elif is_receivearray_class(restype):
            return_length, return_pointer = args
            ReceiveArrayCallback.handle_result(get_args(restype)[0], return_length, return_pointer, result)
        elif restype is WinRT_String:
            return_pointer = args[0]
            return_pointer[0] = WinRT_String(result, own=False)
        elif is_com_class(restype):
            return_pointer = args[0]
            if result:
                result.AddRef()
            return_pointer[0] = result
        else:
            return_pointer = args[0]
            return_pointer[0] = result


class ComClass(ComPtr):
    _keep_reference_in_python_world_ = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._iid_vtbls = self._make_iid_vtbls()
        self.value = addressof(self._iid_vtbls[0][1])  # select default interface
        self._refcount = 0
        self.AddRef()
        self._inner = self._compose()

    def _make_iid_vtbls(self) -> list[tuple[Guid, Vtbl]]:
        iid_vtbls = []
        for interface in self._implemented_interfaces:
            iid = self._get_iid(interface)
            vtbl = Vtbl(self, interface)
            iid_vtbls.append((iid, vtbl))
        return iid_vtbls

    def _get_iid(self, interface):
        if is_generic_alias(interface):
            return _ro_get_parameterized_type_instance_iid(interface)
        elif "_iid_" in interface.__dict__:
            return interface.__dict__["_iid_"]
        else:
            raise ValueError("Cannot get iid")

    @property
    def _implemented_interfaces(self) -> list[_GenericAlias | type]:
        r = []
        for base in self._enumerate_specialized_bases_recursively(_get_original_class(self)):
            if is_generic_alias(base) and "_piid_" in get_origin(base).__dict__:
                r.append(base)
            elif "_iid_" in base.__dict__:
                r.append(base)
        r.append(IAgileObject)
        return r

    def _enumerate_specialized_bases_recursively(self, cls: _GenericAlias | type) -> Iterable[_GenericAlias | type]:
        for base in _get_specialized_bases(cls):
            yield base
            yield from self._enumerate_specialized_bases_recursively(base)

    @classmethod
    def _runtime_class_name(cls) -> str:
        return f"{cls.__module__}.{cls.__qualname__}"

    def _compose(self) -> IInspectable:
        for base in type(self).__mro__:
            # FIXME: ComposableAttribute should be checked.
            if issubclass(base, IInspectable) and "CreateInstance" in base.__dict__:
                inner = IInspectable()
                base.CreateInstance(self, inner)
                return inner
        return None

    def QueryInterface(self, riid: POINTER(Guid), ppvObject: POINTER(VoidPtr)) -> HRESULT:
        for iid, vtbl in self._iid_vtbls:
            if riid[0] == iid:
                ppvObject[0] = addressof(vtbl)
                self.AddRef()
                return S_OK
        if self._inner is not None:
            return self._inner.QueryInterface(riid, ppvObject)
        return E_NOINTERFACE

    def AddRef(self) -> UInt32:
        self._refcount += 1
        if self._refcount == 1:
            self._keep_reference_in_python_world_[id(self)] = self
        return self._refcount

    def Release(self) -> UInt32:
        self._refcount -= 1
        if self._refcount == 0:
            del self._keep_reference_in_python_world_[id(self)]
        return self._refcount

    def GetIids(self, iidCount: POINTER(UInt32), iids: POINTER(POINTER(Guid))) -> HRESULT:
        iidCount[0] = len(self._iid_vtbls)
        iids[0] = CoTaskMemAlloc(sizeof(VoidPtr) * len(self._iid_vtbls))
        if iids[0] is None:
            return E_FAIL
        for i, (iid, vtbl) in enumerate(self._iid_vtbls):
            iids[0][i] = iid
        return S_OK

    def GetRuntimeClassName(self, className: POINTER(HSTRING)) -> HRESULT:
        className[0] = _windows_create_string(self._runtime_class_name())
        return S_OK

    def GetTrustLevel(self, trustLevel: POINTER(TrustLevel)) -> HRESULT:
        trustLevel[0] = BaseTrust
        return S_OK


class MulticastDelegate(IUnknown):
    pass


class MulticastDelegateImpl(ComClass):
    def __init__(self, interface, callback):
        if not callable(callback):
            raise TypeError(f"cannot convert {type(callback)} to MulticastDelegate")
        self._interface = interface
        self._callback = callback
        super().__init__(own=True)

    @property
    def _implemented_interfaces(self):
        return [self._interface, IUnknown, IAgileObject]

    def Invoke(self, *args):
        if inspect.iscoroutinefunction(self._callback):
            self._addref(args)
            future = asyncui.create_task(self._callback(*args))
            future.add_done_callback(lambda _: self._release(args))
            return None
        return self._callback(*args)

    def _addref(self, args):
        for obj in args:
            if isinstance(obj, IUnknown) and obj.value:
                obj.AddRef()

    def _release(self, args):
        for obj in args:
            if isinstance(obj, IUnknown) and obj.value:
                obj.Release()


class AwaitableProtocol:
    def __await__(self):
        future = asyncio.get_running_loop().create_future()
        self.Completed = partial(self.__on_completed, future)
        return future.__await__()

    def __on_completed(self, future, asyncInfo, asyncStatus):
        from win32more.Windows.Foundation import AsyncStatus, IAsyncInfo

        if asyncStatus == AsyncStatus.Completed:
            future.get_loop().call_soon_threadsafe(future.set_result, asyncInfo.GetResults())
        elif asyncStatus == AsyncStatus.Error:
            # It seems that restricted error info is built by invoking IAsyncInfo.ErrorCode.
            error = ComError(asyncInfo.as_(IAsyncInfo).ErrorCode.Value)
            future.get_loop().call_soon_threadsafe(future.set_exception, error)
        elif asyncStatus == AsyncStatus.Canceled:
            future.get_loop().call_soon_threadsafe(future.cancel)
        else:
            assert False, "unreachable"


class IterableProtocol:
    def __class_getitem__(cls, key):
        return cls

    def __iter__(self):
        iterator = self.First()
        while iterator.HasCurrent:
            yield iterator.Current
            iterator.MoveNext()


class SequenceProtocol:
    def __class_getitem__(cls, key):
        return cls

    def __len__(self):
        return self.Size

    def __getitem__(self, index):
        if isinstance(index, slice):
            return [self[i] for i in range(*index.indices(len(self)))]
        elif isinstance(index, int):
            if index < 0:
                index += len(self)
            if index < 0 or len(self) <= index:
                raise IndexError("list index out of range")
            return self.GetAt(index)
        else:
            raise TypeError(f"list indices must be integers or slices, not {type(index).__name__}")


class MappingProtocol:
    def __class_getitem__(cls, key):
        classdict = dict(cls.__dict__)
        classdict["_MappingProtocol__parameters"] = key
        return type("MappingProtocol", (), classdict)

    def __args(self):
        if not is_generic_instance(self):
            return self.__parameters
        gs = GenericSpecializer.from_generic_alias(_get_original_class(self))
        return gs.specialize_parameters(self.__parameters)

    def __iterator(self):
        from win32more.Windows.Foundation.Collections import IIterable, IKeyValuePair

        args = self.__args()
        iterable = self.as_(IIterable[IKeyValuePair[args[0], args[1]]])
        for pair in iterable:
            yield pair.Key, pair.Value

    def __iter__(self):
        yield from self.keys()

    def __len__(self):
        return self.Size

    def __getitem__(self, key):
        if not self.HasKey(key):
            raise KeyError(key)
        return self.Lookup(key)

    def __contains__(self, key):
        return self.HasKey(key)

    def items(self):
        for k, v in self.__iterator():
            yield k, v

    def keys(self):
        for k, v in self.__iterator():
            yield k

    def values(self):
        for k, v in self.__iterator():
            yield v

    def get(self, key, default=None):
        if not self.HasKey(key):
            return default
        return self.Lookup(key)

    def __eq__(self, other):
        return dict(self.items()) == dict(other.items())


# IClosable
class ContextManagerProtocol:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        from win32more.Windows.Foundation import IClosable

        self.as_(IClosable).Close()


class ComError(OSError):
    def __init__(self, hr):
        self._hr = hr
        self._error_info = IErrorInfo(own=True)
        if GetErrorInfo(0, self._error_info) != S_OK:
            self._error_info = None
        super().__init__(hr, self._message())

    def _message(self) -> str:
        return (
            self._winrt_restricted_description()
            or self._winrt_description()
            or self._com_description()
            or self._win32_message()
            or ""
        )

    def _win32_message(self) -> str | None:
        return WinError(self._hr).strerror  # FormatMessageW

    def _com_description(self) -> str | None:
        if self._error_info is None:
            return None

        description = BSTR()
        r = self._error_info.GetDescription(description)
        if r != S_OK:
            return None

        try:
            return self._copy_message(description)
        finally:
            SysFreeString(description)

    def _winrt_description(self) -> str | None:
        details = self._winrt_error_details()
        if details is None:
            return None
        return details["description"]

    def _winrt_restricted_description(self) -> str | None:
        details = self._winrt_error_details()
        if details is None:
            return None
        return details["restricted_description"]

    def _winrt_error_details(self) -> str | None:
        if self._error_info is None:
            return None

        try:
            restricted_error_info = self._error_info.as_(IRestrictedErrorInfo)
        except OSError:
            return None

        description = BSTR()
        error = HRESULT()
        restricted_description = BSTR()
        capability_sid = BSTR()
        r = restricted_error_info.GetErrorDetails(description, error, restricted_description, capability_sid)
        if r != S_OK:
            return None

        try:
            return {
                "restricted_description": self._copy_message(restricted_description),
                "error": error.value,
                "description": self._copy_message(description),
                "capability_sid": self._copy_message(capability_sid),
            }
        finally:
            SysFreeString(description)
            SysFreeString(restricted_description)
            SysFreeString(capability_sid)

    def _copy_message(self, p: BSTR) -> str | None:
        if not p:
            return None
        return wstring_at(p, SysStringLen(p)).strip()



import importlib
import weakref
import xml.etree.ElementTree as ET
from functools import partial
from pathlib import Path
from tempfile import NamedTemporaryFile

from win32more.Microsoft.UI.Xaml import Application, FrameworkElement, IApplicationOverrides
from win32more.Microsoft.UI.Xaml.Markup import IComponentConnector, IXamlMetadataProvider, IXamlType, XamlReader
from win32more.Microsoft.UI.Xaml.XamlTypeInfo import XamlControlsXamlMetaDataProvider
from win32more.Windows.Foundation import Uri
from win32more.Windows.UI.Xaml.Interop import TypeName
from win32more.Windows.Win32.Storage.Packaging.Appx import PACKAGE_VERSION
from win32more.Windows.Win32.System.Com import COINIT_APARTMENTTHREADED, CoInitializeEx, CoUninitialize
from win32more.Windows.Win32.UI.HiDpi import DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2, SetProcessDpiAwarenessContext
from win32more.Windows.Win32.UI.WindowsAndMessaging import SetTimer

from win32more import FAILED, WinError, asyncui
from win32more.appsdk.mddbootstrap import (
    WINDOWSAPPSDK_RELEASE_MAJORMINOR,
    WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W,
    WINDOWSAPPSDK_RUNTIME_VERSION_UINT64,
    MddBootstrapInitialize2,
    MddBootstrapInitializeOptions_OnNoMatch_ShowUI,
    MddBootstrapShutdown,
)
from win32more.winrt import ComClass, WinRT_String

XMLNS_XAML = "http://schemas.microsoft.com/winfx/2006/xaml"
XMLNS_XAML_PRESENTATION = "http://schemas.microsoft.com/winfx/2006/xaml/presentation"

# FIXME: register_namespace() is global.
# This is required to prevent "ns0:" prefix for default namespace.
ET.register_namespace("", XMLNS_XAML_PRESENTATION)


class XamlApplication(ComClass, Application, IApplicationOverrides, IXamlMetadataProvider):
    def __init__(self):
        XamlApplication.__current = self
        self._provider = None
        super().__init__(own=True)
        self.InitializeComponent()
        if WINDOWSAPPSDK_RELEASE_MAJORMINOR == 0x00010008:
            self.ResourceManagerRequested += self._resource_manager_requested

    # Workaround for WindowsAppSDK1.8.0 (microsoft/WindowsAppSDK#5814)
    def _resource_manager_requested(self, sender, e):
        from ctypes import create_unicode_buffer

        from win32more.Microsoft.Windows.ApplicationModel.Resources import ResourceManager
        from win32more.Windows.Win32.System.LibraryLoader import GetModuleFileName, GetModuleHandle

        buf = create_unicode_buffer(1024)
        GetModuleFileName(GetModuleHandle("MRM"), buf, 1024)
        pri = str(Path(buf.value).parent / "resources.pri")
        e.CustomResourceManager = ResourceManager(pri)

    def InitializeComponent(self):
        xaml_path = Path(__file__).with_name("app.xaml").as_posix()
        resource_locator = Uri(f"ms-appx:///{xaml_path}")
        Application.LoadComponent(self, resource_locator)

    def OnLaunched(self, args):
        # You should override this in your derived class
        ...

    def GetXamlType(self, type):
        return self.AppProvider().GetXamlType(type)

    # TODO: Is it needed to provide information for primitive or winui type?
    def GetXamlTypeByFullName(self, fullName):
        return self.AppProvider().GetXamlTypeByFullName(fullName)

    def GetXmlnsDefinitions(self):
        return self.AppProvider().GetXmlnsDefinitions()

    def AppProvider(self):
        if self._provider is None:
            self._provider = XamlControlsXamlMetaDataProvider()
        return self._provider

    __current = None

    @classmethod
    def Start(cls, init):
        r = SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2)
        if not r:
            raise WinError()

        hr = CoInitializeEx(None, COINIT_APARTMENTTHREADED)
        if FAILED(hr):
            raise WinError(hr)

        with asyncui.HandCrankRunner() as runner:
            SetTimer(0, 0, 100, lambda *_: runner.update())
            Application.Start(lambda params: init())

        # FIXME: force Release() to avoid exit with error code.
        if XamlApplication.__current is not None:
            XamlApplication.__current.Release()

        CoUninitialize()


class XamlClass(ComClass, IComponentConnector):
    def __init__(self, *args, own=True, **kwargs):
        super().__init__(*args, own=own, **kwargs)
        self.__component_connector = None

    def Connect(self, connectionId, target):
        self.__component_connector.Connect(connectionId, target)

    def GetBindingConnector(self, connectionId, target):
        return self.__component_connector.GetBindingConnector(connectionId, target)

    def LoadComponentFromFile(self, xaml_path, encoding : str | None = None):
        self.LoadComponentFromString(Path(xaml_path).read_text(encoding=encoding), xaml_path)

    def LoadComponentFromString(self, xaml_str, xaml_path=None):
        self.__component_connector = XamlComponentConnector()
        self.__component_connector.Load(self, xaml_str, xaml_path)


class XamlComponentConnector:
    def __init__(self):
        self._connectors = {}

    def Connect(self, connectionId, target):
        for connect in self._connectors[connectionId]:
            connect(target)

    def GetBindingConnector(self, connectionId, target):
        return None

    def Load(self, component, xaml_str, xaml_path):
        xaml_preprocessed = self._preprocess(component, xaml_str, xaml_path)
        with NamedTemporaryFile(delete_on_close=False) as f:
            f.write(xaml_preprocessed.encode("utf-8"))
            f.close()
            tmp_xaml_path = Path(f.name).as_posix()
            resource_locator = Uri(f"ms-appx:///{tmp_xaml_path}")
            Application.LoadComponent(component, resource_locator)

    def _preprocess(self, component, xaml_str, xaml_path):
        root = ET.fromstring(xaml_str)
        for i, e in enumerate(root.iter()):
            self._connectors[i] = []
            for k, v in list(e.attrib.items()):
                if k == f"{{{XMLNS_XAML}}}Name":
                    self._connectors[i].append(partial(self._connect_name, component, v))
                elif k in _event_names:
                    if e == root:
                        self._connectors[i].append(partial(self._connect_event_root, component, k, v))
                    else:
                        self._connectors[i].append(partial(self._connect_event, component, k, v))
                    del e.attrib[k]
                elif k in _path_names and xaml_path is not None:
                    # FIXME: Workaround for relative path.
                    src_path = Path(xaml_path).parent / v
                    if src_path.exists():
                        e.attrib[k] = src_path.absolute().as_posix()
            if self._connectors[i]:
                e.attrib[f"{{{XMLNS_XAML}}}ConnectionId"] = str(i)
        return ET.tostring(root, encoding="unicode")

    def _connect_name(self, component, bind_name, target):
        setattr(component, bind_name, as_runtime_class(target))

    def _connect_event(self, component, event_name, method_name, target):
        event_setter = getattr(as_runtime_class(target), event_name)
        event_setter += getattr(component, method_name)

    def _connect_event_root(self, component, event_name, method_name, target):
        event_setter = getattr(component, event_name)
        event_setter += getattr(component, method_name)


# Load xaml and connect element and event handler to view object.
#
# <Button x:Name="Button1" Click="Button1_Click" />
#
# to be
#
# view.Button1 = Button()
# view.Button1.Click += view.Button1_Click
class XamlLoader:
    @classmethod
    def Load(cls, view: object, xaml_str: str) -> object:
        return cls().execute(view, xaml_str)

    def __init__(self):
        self._connectors = {}

    def execute(self, view, xaml_str):
        xaml_preprocessed = self._preprocess(view, xaml_str)
        uiroot = as_runtime_class(XamlReader.Load(xaml_preprocessed))
        self._connect(uiroot)
        return uiroot

    def _preprocess(self, view, xaml_str):
        root = ET.fromstring(xaml_str)
        for i, e in enumerate(root.iter()):
            if e is root:
                name = ""
            elif f"{{{XMLNS_XAML}}}Name" in e.attrib:
                name = e.attrib[f"{{{XMLNS_XAML}}}Name"]
            else:
                name = f"__dummy{i}"
            self._connectors[name] = []
            for k, v in list(e.attrib.items()):
                if k == f"{{{XMLNS_XAML}}}Name" and e is not root:
                    self._connectors[name].append(partial(self._connect_name, view, name))
                elif k in _event_names:
                    self._connectors[name].append(partial(self._connect_event, view, k, v))
                    del e.attrib[k]
            if self._connectors[name] and f"{{{XMLNS_XAML}}}Name" not in e.attrib and e is not root:
                e.attrib[f"{{{XMLNS_XAML}}}Name"] = name
        return ET.tostring(root, encoding="unicode")

    def _connect(self, uiroot):
        try:
            framework_element = uiroot.as_(FrameworkElement)
        except OSError:
            framework_element = uiroot.Content.as_(FrameworkElement)

        for name, connectors in self._connectors.items():
            for connect in connectors:
                if name == "":
                    target = uiroot
                else:
                    target = framework_element.FindName(name)
                connect(target)

    def _connect_name(self, view, bind_name, target):
        setattr(view, bind_name, as_runtime_class(target))

    def _connect_event(self, view, event_name, method_name, target):
        event_setter = getattr(as_runtime_class(target), event_name)
        event_setter += getattr(view, method_name)


class XamlType(ComClass, IXamlType):
    def __init__(
        self,
        full_name,
        type_kind,
        *,
        base_type=None,
        boxed_type=None,
        content_property=None,
        is_array=False,
        is_bindable=False,
        is_collection=False,
        is_dictionary=False,
        is_markup_extension=False,
        item_type=None,
        key_type=None,
        activate_instance=None,
        add_to_map=None,
        add_to_vector=None,
        create_from_string=None,
    ):
        super().__init__(own=True)
        self._full_name = full_name
        self._type_kind = type_kind
        self._base_type = base_type
        self._boxed_type = boxed_type
        self._content_property = content_property
        self._is_array = is_array
        self._is_bindable = is_bindable
        self._is_collection = is_collection
        self._is_constructible = bool(activate_instance)
        self._is_dictionary = is_dictionary
        self._is_markup_extension = is_markup_extension
        self._item_type = item_type
        self._key_type = key_type
        self._activate_instance = activate_instance
        self._add_to_map = add_to_map
        self._add_to_vector = add_to_vector
        self._create_from_string = create_from_string

    def get_BaseType(self):
        return self._base_type

    def get_BoxedType(self):
        return self._boxed_type

    def get_ContentProperty(self):
        return self._content_property

    def get_FullName(self):
        return self._full_name

    def get_IsArray(self):
        return self._is_array

    def get_IsBindable(self):
        return self._is_bindable

    def get_IsCollection(self):
        return self._is_collection

    def get_IsConstructible(self):
        return self._is_constructible

    def get_IsDictionary(self):
        return self._is_dictionary

    def get_IsMarkupExtension(self):
        return self._is_markup_extension

    def get_ItemType(self):
        return self._item_type

    def get_KeyType(self):
        return self._key_type

    def get_UnderlyingType(self):
        return TypeName(WinRT_String(self._full_name), self._type_kind)

    def ActivateInstance(self):
        if self._activate_instance is None:
            raise NotImplementedError()
        return self._activate_instance()

    def AddToMap(self, instance, key, value):
        if self._add_to_map is None:
            raise NotImplementedError()
        self._add_to_map(instance, key, value)

    def AddToVector(self, instance, value):
        if self._add_to_vector is None:
            raise NotImplementedError()
        self._add_to_vector(instance, value)

    def CreateFromString(self, value):
        if self._boxed_type is not None:
            return self._boxed_type.CreateFromString(value)
        if self._create_from_string is not None:
            return self._create_from_string(value)
        # TODO:
        # return fromStringConverter()
        raise NotImplementedError()

    def GetMember(self, name):
        # TODO:
        raise NotImplementedError()

    def RunInitializer(self):
        pass


def xaml_typename(name, kind):
    hs = WinRT_String(name)
    tn = TypeName(hs, kind)
    weakref.finalize(tn, hs.clear)
    return tn


def as_runtime_class(uielement):
    return uielement.as_(_get_runtime_class(uielement))


def _get_runtime_class(uielement):
    namespace, name = _get_runtime_class_name(uielement).rsplit(".", 1)
    module = importlib.import_module(f"win32more.{namespace}")
    return getattr(module, name)


def _get_runtime_class_name(uielement):
    s = WinRT_String(own=True)
    hr = uielement.GetRuntimeClassName(s)
    if FAILED(hr):
        raise WinError(hr)
    return s.strvalue


# FIXME: how to detect event name?
_event_names = {
    "AccessKeyDisplayDismissed",
    "AccessKeyDisplayRequested",
    "AccessKeyInvoked",
    "ActionButtonClick",
    "Activated",
    "ActualPlacementChanged",
    "ActualThemeChanged",
    "AddPages",
    "AddScrollVelocityRequested",
    "AddTabButtonClick",
    "AnchorRequested",
    "AnimatedVisualInvalidated",
    "ArrangeInvalidated",
    "BackRequested",
    "BeforeTextChanging",
    "BindingFailed",
    "BringIntoViewRequested",
    "BringingIntoView",
    "CalendarViewDayItemChanging",
    "CanExecuteChanged",
    "CanExecuteRequested",
    "CanScrollChanged",
    "CandidateWindowBoundsChanged",
    "Changed",
    "CharacterReceived",
    "Checked",
    "ChoosingGroupHeaderContainer",
    "ChoosingItemContainer",
    "CleanUpVirtualizedItemEvent",
    "Click",
    "CloseButtonClick",
    "CloseRequested",
    "Closed",
    "Closing",
    "Collapsed",
    "CollectionChanged",
    "ColorChanged",
    "Completed",
    "CompositionScaleChanged",
    "Confirmed",
    "ContainerContentChanging",
    "ContextCanceled",
    "ContextMenuOpening",
    "ContextRequested",
    "CopyingToClipboard",
    "CoreProcessFailed",
    "CoreWebView2Initialized",
    "CurrentChanged",
    "CurrentChanging",
    "CurrentStateChanged",
    "CurrentStateChanging",
    "CuttingToClipboard",
    "DataContextChanged",
    "DateChanged",
    "DatePicked",
    "DetailLabelRequested",
    "DirectManipulationCompleted",
    "DirectManipulationStarted",
    "DisplayModeChanged",
    "DoubleTapped",
    "DownloadProgress",
    "DragCompleted",
    "DragDelta",
    "DragEnter",
    "DragItemsCompleted",
    "DragItemsStarting",
    "DragLeave",
    "DragOver",
    "DragStarted",
    "DragStarting",
    "Drop",
    "DropCompleted",
    "DropDownClosed",
    "DropDownOpened",
    "DynamicOverflowItemsChanging",
    "EffectiveViewportChanged",
    "ElementClearing",
    "ElementIndexChanged",
    "ElementPrepared",
    "ErrorsChanged",
    "ExecuteRequested",
    "Expanding",
    "ExtentChanged",
    "ExternalTornOutTabsDropped",
    "ExternalTornOutTabsDropping",
    "FocusDisengaged",
    "FocusEngaged",
    "GetPreviewPage",
    "GettingFocus",
    "GotFocus",
    "Holding",
    "HorizontalSnapPointsChanged",
    "ImageFailed",
    "ImageOpened",
    "Indeterminate",
    "Invoked",
    "IsCheckedChanged",
    "IsDisplayModeEnabledChanged",
    "IsEnabledChanged",
    "IsScrollingWithMouseChanged",
    "IsTextTrimmedChanged",
    "ItemClick",
    "ItemClicked",
    "ItemInvoked",
    "ItemsChanged",
    "ItemsInfoRequested",
    "ItemsPicked",
    "ItemsUnlocked",
    "KeyDown",
    "KeyUp",
    "LayoutUpdated",
    "LoadCompleted",
    "Loaded",
    "Loading",
    "LosingFocus",
    "LostFocus",
    "ManipulationCompleted",
    "ManipulationDelta",
    "ManipulationInertiaStarting",
    "ManipulationStarted",
    "ManipulationStarting",
    "MapElementClick",
    "MapServiceErrorOccurred",
    "MeasureInvalidated",
    "ModeChanged",
    "Navigated",
    "Navigating",
    "NavigationCompleted",
    "NavigationFailed",
    "NavigationStarting",
    "NavigationStopped",
    "NoFocusCandidateFound",
    "OpenFailed",
    "Opened",
    "Opening",
    "Paginate",
    "PanRequested",
    "PaneClosed",
    "PaneClosing",
    "PaneOpened",
    "PaneOpening",
    "PasswordChanged",
    "PasswordChanging",
    "Paste",
    "PivotItemLoaded",
    "PivotItemLoading",
    "PivotItemUnloaded",
    "PivotItemUnloading",
    "PointerCanceled",
    "PointerCaptureLost",
    "PointerEntered",
    "PointerExited",
    "PointerMoved",
    "PointerPressed",
    "PointerReleased",
    "PointerWheelChanged",
    "PreviewKeyDown",
    "PreviewKeyUp",
    "PrimaryButtonClick",
    "ProcessKeyboardAccelerators",
    "PropertyChanged",
    "QuerySubmitted",
    "RefreshRequested",
    "RefreshStateChanged",
    "Rendered",
    "Rendering",
    "ResourceManagerRequested",
    "RightTapped",
    "Scroll",
    "ScrollAnimationStarting",
    "ScrollByRequested",
    "ScrollCompleted",
    "ScrollToRequested",
    "Scrolling",
    "SecondaryButtonClick",
    "SectionHeaderClick",
    "SectionsInViewChanged",
    "SelectedDateChanged",
    "SelectedDatesChanged",
    "SelectedIndexChanged",
    "SelectedTimeChanged",
    "SelectionChanged",
    "SelectionChanging",
    "SizeChanged",
    "StateChanged",
    "SuggestionChosen",
    "SurfaceContentsLost",
    "TabCloseRequested",
    "TabDragCompleted",
    "TabDragStarting",
    "TabDroppedOutside",
    "TabItemsChanged",
    "TabStripDragOver",
    "TabStripDrop",
    "TabTearOutRequested",
    "TabTearOutWindowRequested",
    "TakeFocusRequested",
    "Tapped",
    "TextChanged",
    "TextChanging",
    "TextCompositionChanged",
    "TextCompositionEnded",
    "TextCompositionStarted",
    "TextSubmitted",
    "ThumbnailRequested",
    "Tick",
    "TimeChanged",
    "TimePicked",
    "Toggled",
    "TransitionCompleted",
    "Unchecked",
    "UnhandledException",
    "Unloaded",
    "ValueChanged",
    "VectorChanged",
    "VerticalSnapPointsChanged",
    "ViewChangeCompleted",
    "ViewChangeStarted",
    "ViewChanged",
    "ViewChanging",
    "VisibilityChanged",
    "WebMessageReceived",
    "XamlResourceReferenceFailed",
    "XamlShutdownCompletedOnThread",
    "ZoomAnimationStarting",
    "ZoomCompleted",
}


_path_names = {
    "Source",
    "ProfilePicture",
}



# ruff: noqa: F401
from win32more import (
    ARCH,
    FAILED,
    POINTER,
    SUCCEEDED,
    Annotated,
    Boolean,
    Byte,
    Bytes,
    Char,
    ComPtr,
    ConstantLazyLoader,
    Double,
    Enum,
    FlexibleArray,
    Guid,
    Int16,
    Int32,
    Int64,
    IntPtr,
    NativeBitfieldAttribute,
    SByte,
    Single,
    String,
    Structure,
    UInt16,
    UInt32,
    UInt64,
    UIntPtr,
    UnicodeAlias,
    Union,
    Void,
    VoidPtr,
    cfunctype,
    cfunctype_pointer,
    commethod,
    make_ready,
    winfunctype,
    winfunctype_pointer,
)



import asyncio
import sys
import threading
from concurrent.futures import Future, ThreadPoolExecutor


class HandCrankRunner:
    def __init__(self, *, loop_factory=None):
        if loop_factory is None:
            self._loop_factory = maybe_eager_loop_factory
        else:
            self._loop_factory = loop_factory
        self._loop = None

    def __enter__(self):
        self._loop = self._loop_factory()
        self._loop.stop()
        if sys.version_info < (3, 13):
            asyncio.events._set_running_loop(self._loop)
        else:
            self._loop._run_forever_setup()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if sys.version_info < (3, 13):
            asyncio.events._set_running_loop(None)
        else:
            self._loop._run_forever_cleanup()
        self._loop.close()
        self._loop = None

    def update(self):
        self._loop._run_once()


def create_task(coro):
    try:
        executor = RunningLoopTaskExecutor()
    except RuntimeError:
        executor = ThreadPoolTaskExecutor(maybe_eager_loop_factory)
    future = executor.submit(coro)
    BackgroundTasks.add(future)
    future.add_done_callback(BackgroundTasks.discard)
    return future


def maybe_eager_loop_factory():
    loop = asyncio.new_event_loop()
    # Use eager task.
    # Some method can not be called after returned.
    # (e.g. CoreWebView2NewWindowRequestedEventArgs.GetDeferral())
    if sys.version_info >= (3, 12):
        loop.set_task_factory(asyncio.eager_task_factory)
    return loop


# keep reference to tasks
class BackgroundTasks:
    _tasks = set()
    _lock = threading.Lock()

    @classmethod
    def add(cls, task):
        with cls._lock:
            cls._tasks.add(task)

    @classmethod
    def discard(cls, task):
        with cls._lock:
            cls._tasks.discard(task)


class RunningLoopTaskExecutor:
    def __init__(self):
        self._loop = asyncio.get_running_loop()

    def submit(self, coro):
        return self._loop.create_task(coro)


class ThreadPoolTaskExecutor:
    _thread_pool = None
    _lock = threading.Lock()

    def __init__(self, loop_factory):
        self._loop_factory = loop_factory
        self._init_thread_pool()

    @classmethod
    def _init_thread_pool(cls):
        with cls._lock:
            if cls._thread_pool is None:
                cls._thread_pool = ThreadPoolExecutor()

    def submit(self, coro):
        loop = self._loop_factory()
        # Run task in current thread context until first await.
        # eager_task_factory doesn't start task eagerly when loop is not running.
        task = loop.create_task(coro)
        loop.stop()
        loop.run_forever()
        if task.done():
            loop.run_until_complete(task)  # ensure calling done callback
            loop.close()
            future = Future()
            future.set_result(task.result())
            return future
        # continue in background thread
        return self._thread_pool.submit(self._background, loop, task)

    def _background(self, loop, task):
        loop.run_until_complete(task)
        loop.close()
        return task.result()



from __future__ import annotations

import logging

from win32more import POINTER, Boolean, UInt32, Void
from win32more.Windows.Foundation import EventRegistrationToken
from win32more.Windows.Foundation.Collections import (
    CollectionChange,
    IIterable,
    IIterator,
    IObservableVector,
    IVector,
    IVectorChangedEventArgs,
    IVectorView,
    VectorChangedEventHandler,
)
from win32more.winrt import (
    ComClass,
    FillArray,
    PassArray,
    T,
    get_args,
    is_com_class,
)

logger = logging.getLogger(__name__)


class call_after_orig_class_set:
    def __init__(self, orig_init):
        self._orig_init = orig_init

    def __set_name__(self, owner, name):
        owner.__orig_class__ = self

    def __call__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs

    def __set__(self, instance, value):
        instance.__dict__["__orig_class__"] = value
        try:
            self._orig_init(instance, *self._args, **self._kwargs)
        except Exception:
            # _BaseGenericAlias.__call__() ignore error
            logger.exception("Unhandled exception caught")
            raise


class Vector(ComClass, IVector[T], IVectorView[T], IIterable[T], IObservableVector[T]):
    @call_after_orig_class_set
    def __init__(self, lst: list[T] | None = None) -> None:
        super().__init__(own=True)

        self._type = get_args(self.__orig_class__)[0]

        self._observers = {}
        self._observers_count = 0

        if lst is None:
            self._lst = []
        else:
            self._lst = list(lst)

        if is_com_class(self._type):
            for v in self._lst:
                v.AddRef()

    def GetAt(self, index: UInt32) -> T:
        return self._lst[index]

    def get_Size(self) -> UInt32:
        return len(self._lst)

    def GetView(self) -> IVectorView[T]:
        return self.as_(IVector[self._type])

    def IndexOf(self, value: T, index: POINTER(UInt32)) -> Boolean:
        for i, v in enumerate(self._lst):
            if v == value:
                index[0] = i
                return True
        return False

    def SetAt(self, index: UInt32, value: T) -> Void:
        if is_com_class(self._type):
            self._lst[index].Release()
            value.AddRef()
        self._lst[index] = value
        self._notify(VectorChangedEventArgs(CollectionChange.ItemChanged, index))

    def InsertAt(self, index: UInt32, value: T) -> Void:
        if is_com_class(self._type):
            value.AddRef()
        self._lst.insert(index, value)
        self._notify(VectorChangedEventArgs(CollectionChange.ItemInserted, index))

    def RemoveAt(self, index: UInt32) -> Void:
        value = self._lst.pop(index)
        if is_com_class(self._type):
            value.Release()
        self._notify(VectorChangedEventArgs(CollectionChange.ItemRemoved, index))

    def Append(self, value: T) -> Void:
        self.InsertAt(len(self._lst), value)

    def RemoveAtEnd(self) -> Void:
        self.RemoveAt(len(self) - 1)

    def Clear(self) -> Void:
        if is_com_class(self._type):
            for v in self._lst:
                v.Release()
        self._lst[:] = []
        self._notify(VectorChangedEventArgs(CollectionChange.Reset, 0))

    def GetMany(self, startIndex: UInt32, items: FillArray[T]) -> UInt32:
        numcopied = 0
        for i in range(startIndex, startIndex + len(items)):
            if i >= len(self._lst):
                break
            items[numcopied] = self._lst[i]
            numcopied += 1
        return numcopied

    def ReplaceAll(self, items: PassArray[T]) -> Void:
        if is_com_class(self._type):
            for v in items:
                v.AddRef()
        if is_com_class(self._type):
            for v in self._lst:
                v.Release()
        self._lst[:] = items
        # FIXME: ?
        self._notify(VectorChangedEventArgs(CollectionChange.Reset, 0))

    def First(self) -> IIterator[T]:
        return Iterator[self._type](self._lst)

    def add_VectorChanged(self, vhnd: VectorChangedEventHandler[T]) -> EventRegistrationToken:
        self._observers_count += 1
        self._observers[self._observers_count] = vhnd
        vhnd.AddRef()
        return EventRegistrationToken(self._observers_count)

    def remove_VectorChanged(self, token: EventRegistrationToken) -> Void:
        vhnd = self._observers.pop(token.Value)
        vhnd.Release()

    def _notify(self, args):
        for observer in self._observers.values():
            observer.Invoke(self, args)


class Iterator(ComClass, IIterator[T]):
    @call_after_orig_class_set
    def __init__(self, lst: list[T] | None = None) -> None:
        super().__init__(own=True)

        self._type = get_args(self.__orig_class__)[0]

        self._lst = lst
        self._index = 0

    def get_Current(self) -> T:
        return self._lst[self._index]

    def get_HasCurrent(self) -> Boolean:
        return self._index < len(self._lst)

    def MoveNext(self) -> Boolean:
        self._index += 1
        return self._index < len(self._lst)

    def GetMany(self, items: FillArray[T]) -> UInt32:
        numcopied = 0
        for i in range(len(items)):
            if i >= len(self._lst):
                break
            items[numcopied] = self._lst[i]
            numcopied += 1
        return numcopied


class VectorChangedEventArgs(ComClass, IVectorChangedEventArgs):
    def __init__(self, collection_change: CollectionChange, index: int) -> None:
        super().__init__(own=True)
        self._collection_change = collection_change
        self._index = index

    def get_CollectionChange(self) -> CollectionChange:
        return self._collection_change

    def get_Index(self) -> UInt32:
        return self._index



import platform
from ctypes import CFUNCTYPE, POINTER, c_char_p, c_uint, c_void_p, c_wchar_p, cast, cdll, pointer

if platform.machine() != "x86_64":
    raise RuntimeError(f"{platform.machine()} is not supported")

ARCH = "X64"

windll = cdll


def MAKELANGID(p, s):
    return (s << 10) | p


def FormatError(code):
    from win32more.Windows.Win32.Foundation import LocalFree
    from win32more.Windows.Win32.System.Diagnostics.Debug import (
        FORMAT_MESSAGE_ALLOCATE_BUFFER,
        FORMAT_MESSAGE_FROM_SYSTEM,
        FORMAT_MESSAGE_IGNORE_INSERTS,
        FormatMessageW,
    )
    from win32more.Windows.Win32.System.SystemServices import LANG_NEUTRAL, SUBLANG_NEUTRAL

    lpMsgBuf = c_wchar_p()
    n = FormatMessageW(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        None,
        code,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
        cast(pointer(lpMsgBuf), c_wchar_p),
        0,
        None,
    )
    if n == 0:
        return "<no description>"
    msg = lpMsgBuf.value.rstrip()
    LocalFree(lpMsgBuf)
    return msg


def WinError(code=None, descr=None):
    from win32more.Windows.Win32.Foundation import GetLastError

    if code is None:
        code = GetLastError()
    if descr is None:
        descr = FormatError(code).strip()
    # Cygwin version doesn't have 4th winerror argument.
    # OSError(None, descr, None, code)
    return OSError(code, descr)


def WINFUNCTYPE(restype, *argtypes, use_errno=False, use_last_error=False):
    def __new__(cls, *args):
        if len(args) >= 2 and isinstance(args[0], int):
            return _ComMethod(restype, argtypes, use_errno, use_last_error, *args)
        return functype_new(cls, *args)

    functype = CFUNCTYPE(restype, *argtypes, use_errno=use_errno, use_last_error=use_last_error)
    functype_new = functype.__new__
    functype.__new__ = __new__
    return functype


class _ComMethod:
    def __init__(self, restype, argtypes, use_errno, use_last_error, vtbl_index, name, paramflags=None, iid=None):
        self._vtbl_index = vtbl_index
        self._functype = CFUNCTYPE(restype, c_void_p, *argtypes, use_errno=use_errno, use_last_error=use_last_error)

    def __call__(self, this, *args, **kwargs):
        pvtbl = cast(this, POINTER(POINTER(c_void_p)))
        func = self._functype(pvtbl[0][self._vtbl_index])
        # WORKAROUND: return_length and return are passed as kwargs.
        _args = list(args) + list(kwargs.values())
        return func(this, *_args)


CCP_POSIX_TO_WIN_W = 1

try:
    cygwin1 = cdll.LoadLibrary("cygwin1.dll")
except OSError:
    cygwin1 = cdll.LoadLibrary("msys-2.0.dll")

cygwin_create_path = cygwin1.cygwin_create_path
cygwin_create_path.restype = c_void_p
cygwin_create_path.argtypes = [c_uint, c_void_p]

free = cygwin1.free
free.restype = None
free.argtypes = [c_void_p]


def posix_to_win(posix_path: str) -> str:
    p = cygwin_create_path(CCP_POSIX_TO_WIN_W, c_char_p(posix_path.encode("utf-8")))
    win_path = c_wchar_p(p).value
    free(p)
    return win_path



from __future__ import annotations

import os
import sys
import webbrowser
from pathlib import Path

import win32more.Windows.Win32.Foundation
import win32more.Windows.Win32.Storage.Packaging.Appx
from win32more.Windows.Win32.Foundation import (
    APPMODEL_ERROR_NO_PACKAGE,
    ERROR_INSUFFICIENT_BUFFER,
    ERROR_SUCCESS,
    S_OK,
    STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED,
)
from win32more.Windows.Win32.Storage.Packaging.Appx import (
    PACKAGE_VERSION,
    PACKAGEDEPENDENCY_CONTEXT,
    AddPackageDependency,
    AddPackageDependencyOptions_None,
    CreatePackageDependencyOptions_None,
    GetCurrentPackageFullName,
    GetPackagePathByFullName,
    GetPackagesByPackageFamily,
    PackageDependencyLifetimeKind_Process,
    PackageDependencyProcessorArchitectures_Arm64,
    PackageDependencyProcessorArchitectures_X64,
    PackageDependencyProcessorArchitectures_X86,
    TryCreatePackageDependency,
)
from win32more.Windows.Win32.UI.WindowsAndMessaging import IDYES, MB_ICONERROR, MB_YESNO, MessageBox

import win32more.appsdk.mddbootstrap
from win32more import ARCH, FAILED, Char, Int32, String, UInt32, Void, WinError, make_ready, winfunctype
from win32more.winrt import ComError

# versioninfo.py will be installed by win32more-Microsoft.WindowsAppSDK
from .versioninfo import (  # noqa
    WINDOWSAPPSDK_RELEASE_MAJORMINOR,
    WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W,
    WINDOWSAPPSDK_RUNTIME_VERSION_UINT64,
)

_module = sys.modules[__name__]

if ARCH == "ARM64":
    PackageDependencyProcessorArchitectures_Current = PackageDependencyProcessorArchitectures_Arm64
    os.add_dll_directory(os.path.dirname(win32more.__file__) + "\\dll\\arm64")
elif ARCH == "X64":
    PackageDependencyProcessorArchitectures_Current = PackageDependencyProcessorArchitectures_X64
    os.add_dll_directory(os.path.dirname(win32more.__file__) + "\\dll\\x64")
elif ARCH == "X86":
    PackageDependencyProcessorArchitectures_Current = PackageDependencyProcessorArchitectures_X86
    os.add_dll_directory(os.path.dirname(win32more.__file__) + "\\dll\\x86")
else:
    assert False

MddBootstrapInitializeOptions = Int32
MddBootstrapInitializeOptions_None: win32more.appsdk.mddbootstrap.MddBootstrapInitializeOptions = 0
MddBootstrapInitializeOptions_OnError_DebugBreak: win32more.appsdk.mddbootstrap.MddBootstrapInitializeOptions = 1
MddBootstrapInitializeOptions_OnError_DebugBreak_IfDebuggerAttached: win32more.appsdk.mddbootstrap.MddBootstrapInitializeOptions = 2
MddBootstrapInitializeOptions_OnError_FailFast: win32more.appsdk.mddbootstrap.MddBootstrapInitializeOptions = 4
MddBootstrapInitializeOptions_OnNoMatch_ShowUI: win32more.appsdk.mddbootstrap.MddBootstrapInitializeOptions = 8
MddBootstrapInitializeOptions_OnPackageIdentity_NOOP: win32more.appsdk.mddbootstrap.MddBootstrapInitializeOptions = 16


@winfunctype("Microsoft.WindowsAppRuntime.Bootstrap.dll", entry_point="MddBootstrapInitialize")
def _MddBootstrapInitialize(
    majorMinorVersion: UInt32,
    versionTag: String,
    minVersion: win32more.Windows.Win32.Storage.Packaging.Appx.PACKAGE_VERSION,
) -> win32more.Windows.Win32.Foundation.HRESULT: ...


@winfunctype("Microsoft.WindowsAppRuntime.Bootstrap.dll", entry_point="MddBootstrapInitialize2")
def _MddBootstrapInitialize2(
    majorMinorVersion: UInt32,
    versionTag: String,
    minVersion: win32more.Windows.Win32.Storage.Packaging.Appx.PACKAGE_VERSION,
    options: win32more.appsdk.mddbootstrap.MddBootstrapInitializeOptions,
) -> win32more.Windows.Win32.Foundation.HRESULT: ...


@winfunctype("Microsoft.WindowsAppRuntime.Bootstrap.dll", entry_point="MddBootstrapShutdown")
def _MddBootstrapShutdown() -> Void: ...


def MddBootstrapInitialize(major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION) -> int:
    return MddBootstrapInitialize2(major_minor_version, version_tag, min_version, MddBootstrapInitializeOptions_None)


def MddBootstrapInitialize2(
    major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION, options: int
) -> int:
    if _IsPackagedProcess():
        # FIXME: Mddbootstrap API doesn't support packaged process.
        # Since 1.6.5, WindowsAppSDK seems to work with Win11's package dependency API.
        # I'm not sure if it is offically supported.  It might have a problem.
        VERSION_1_6_5 = 0x1770019109300000
        if not _IsWin11() or WINDOWSAPPSDK_RUNTIME_VERSION_UINT64 < VERSION_1_6_5:
            raise RuntimeError("Packaged process is not supported before Windows 11")
        hr = _Initialize_Win11(_GetFrameworkPackageFamilyName(major_minor_version, version_tag), min_version)
        if FAILED(hr):
            if hr == STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED:
                if options & MddBootstrapInitializeOptions_OnNoMatch_ShowUI:
                    _OnNoMatch_ShowUI(major_minor_version, version_tag, min_version)
            return hr
        return S_OK
    else:
        return _module._MddBootstrapInitialize2(major_minor_version, version_tag, min_version, options)


def MddBootstrapShutdown() -> None:
    if _IsPackagedProcess() and not _IsWin11():
        raise RuntimeError("Packaged process is not supported before Windows 11")
    elif _IsWin11():
        pass
    else:
        _module._MddBootstrapShutdown()


def _IsWin11() -> bool:
    return sys.getwindowsversion() >= (10, 0, 22000)


def _IsPackagedProcess() -> bool:
    return GetCurrentPackageFullName(UInt32(), None) != APPMODEL_ERROR_NO_PACKAGE


def _Initialize_Win11(family_name: str, min_version: PACKAGE_VERSION) -> int:
    package_dependency_id = String()
    hr = TryCreatePackageDependency(
        None,
        family_name,
        min_version,
        PackageDependencyProcessorArchitectures_Current,
        PackageDependencyLifetimeKind_Process,
        None,
        CreatePackageDependencyOptions_None,
        package_dependency_id,
    )
    if FAILED(hr):
        return hr

    MDD_PACKAGE_DEPENDENCY_RANK_DEFAULT = 0
    package_dependency_context = PACKAGEDEPENDENCY_CONTEXT()
    hr = AddPackageDependency(
        package_dependency_id,
        MDD_PACKAGE_DEPENDENCY_RANK_DEFAULT,
        AddPackageDependencyOptions_None,
        package_dependency_context,
        None,
    )
    if FAILED(hr):
        return hr

    return S_OK


# https://learn.microsoft.com/en-us/windows/apps/desktop/modernize/package-identity-overview
def _GetFrameworkPackageFamilyName(major_minor_version: int, version_tag: PACKAGE_VERSION) -> str:
    major_version = major_minor_version >> 16
    minor_version = major_minor_version & 0xFFFF
    version_tag_delimiter = "" if version_tag == "" else "-"
    name = f"Microsoft.WindowsAppRuntime.{major_version}.{minor_version}{version_tag_delimiter}{version_tag}"
    publisher_id = "8wekyb3d8bbwe"
    return f"{name}_{publisher_id}"


def _GetFrameworkPackageDirectory(
    major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION
) -> str | None:
    family_name = _GetFrameworkPackageFamilyName(major_minor_version, version_tag)
    full_name = _ResolvePackageDependency(family_name, min_version, ARCH)
    if full_name is None:
        return None
    return _GetPackagePathByFullName(full_name)


def _ResolvePackageDependency(family_name: str, min_version: PACKAGE_VERSION, arch: str) -> str | None:
    bestfit = None
    latest_version = None

    for full_name in _GetPackagesByPackageFamily(family_name):
        name, version, architecture, resource_id, publisher_id = full_name.split("_")
        major, minor, build, revision = [int(x) for x in version.split(".")]
        if architecture.upper() != arch:
            continue
        if (major, minor, build, revision) < (
            min_version.Major,
            min_version.Minor,
            min_version.Build,
            min_version.Revision,
        ):
            continue
        if latest_version is None or latest_version < (major, minor, build, revision):
            bestfit = full_name
            latest_version = (major, minor, build, revision)

    return bestfit


def _OnNoMatch_ShowUI(major_minor_version: int, version_tag: str, min_version: PACKAGE_VERSION) -> None:
    caption = f"{Path(sys.executable).name} - This application could not be started"
    text = f"""This application requires the Windows App Runtime
Version {major_minor_version >> 16}.{major_minor_version & 0xFFFF}
(MSIX package version >= {min_version.Major}.{min_version.Minor}.{min_version.Build}.{min_version.Revision})

Do you want to install a compatible Windows App Runtime now?"""
    r = MessageBox(0, text, caption, MB_YESNO | MB_ICONERROR)
    if r == IDYES:
        webbrowser.open("https://docs.microsoft.com/windows/apps/windows-app-sdk/downloads")


def _GetPackagesByPackageFamily(family_name: str) -> list[str]:
    count = UInt32()
    buffer_length = UInt32()
    r = GetPackagesByPackageFamily(family_name, count, None, buffer_length, None)
    if r == ERROR_SUCCESS:
        return []  # not found
    elif r != ERROR_INSUFFICIENT_BUFFER:
        raise WinError(r)

    full_names = (String * count.value)()
    buffer = (Char * buffer_length.value)()
    r = GetPackagesByPackageFamily(family_name, count, full_names, buffer_length, buffer)
    if r != ERROR_SUCCESS:
        raise WinError(r)

    return full_names[: count.value]


def _GetPackagePathByFullName(full_name: str) -> str:
    path_length = UInt32()
    r = GetPackagePathByFullName(full_name, path_length, None)
    if r != ERROR_INSUFFICIENT_BUFFER:
        raise WinError(r)

    path = (Char * path_length.value)()
    r = GetPackagePathByFullName(full_name, path_length, path)
    if r != ERROR_SUCCESS:
        raise WinError(r)

    return path.value


def initialize() -> None:
    hr = MddBootstrapInitialize(
        WINDOWSAPPSDK_RELEASE_MAJORMINOR,
        WINDOWSAPPSDK_RELEASE_VERSION_SHORTTAG_W,
        PACKAGE_VERSION(Version=WINDOWSAPPSDK_RUNTIME_VERSION_UINT64),
    )

    if FAILED(hr):
        if hr == STATEREPOSITORY_E_DEPENDENCY_NOT_RESOLVED:
            major = WINDOWSAPPSDK_RELEASE_MAJORMINOR >> 16
            minor = WINDOWSAPPSDK_RELEASE_MAJORMINOR & 0xFFFF
            min_version = PACKAGE_VERSION(Version=WINDOWSAPPSDK_RUNTIME_VERSION_UINT64)
            raise OSError(
                hr,
                f"This application requires the Windows App Runtime Version {major}.{minor} (MSIX package version >= {min_version.Major}.{min_version.Minor}.{min_version.Build}.{min_version.Revision})",
            )
        raise ComError(hr)


def uninitialize() -> None:
    MddBootstrapShutdown()


make_ready(__name__)
